import os
import sys
import json
import fnmatch
from typing import List, Optional

from hpogrid.core.defaults import kDefultGridSiteType, kDefaultGridSiteInfo
from hpogrid.utils.stylus import create_table

class GridSiteInfo():
    
    @staticmethod
    def show(name:str=None, gpu:bool=True, active:bool=True, site_type:Optional[List]=kDefultGridSiteType,
             info:List=kDefaultGridSiteInfo):

        grid_site_info = GridSiteInfo.extract(name, gpu, active, site_type, info)
        print(create_table(grid_site_info, transpose=True))

    @staticmethod
    def list_sites(name:str=None, gpu:bool=True, active:bool=True, site_type:Optional[List]=kDefultGridSiteType):
        grid_site_info = GridSiteInfo.extract(name, gpu, active, site_type)
        return list(grid_site_info.keys())

    @staticmethod
    def extract(name:str=None, gpu:bool=True, active:bool=True, site_type:Optional[List]=kDefultGridSiteType,
        info:List=kDefaultGridSiteInfo):
        '''
        retrieve some basic information of PanDA grid sites
        '''
        try:
            jsonfileLocation = os.environ['ALRB_cvmfs_repo'] + '/sw/local/etc/agis_schedconf.json'
        except:
            jsonfileLocation = '/cvmfs/atlas.cern.ch/repo/sw/local/etc/agis_schedconf.json'
        
        if not os.path.exists(jsonfileLocation):
            raise ValueError('cannot locate file containing grid site information: agis_schedconf.json')
            
        with open(jsonfileLocation,'r') as jsonfile:
            jsondata = json.load(jsonfile)

        if name is None:
            name = '*'

        grid_site_info = {}

        for site in jsondata:
            # filter site names (could also match jsondata[site]['panda_resource'] instead)
            if not fnmatch.fnmatch(site, name):
                continue
            # filter non-active grid sites
            if active and (not jsondata[site]['state'] == 'ACTIVE'):
                continue
            # no good indicator of a GPU site yet will just judge on site name
            if gpu and (not 'GPU' in jsondata[site]['panda_resource']):
                continue
            if site_type and (jsondata[site]['type'] not in site_type):
                continue
            grid_site_info[site] = {key: jsondata[site][key] for key in info}

        return grid_site_info


import os
import sys
import copy
import json

import pandas as pd
from tabulate import tabulate

from hpogrid.core.defaults import *
from .validation import validate_job_metadata

kSupportedExtraColumns = ['site', 'task_time_s', 'time_s', 'taskid', 'start_timestamp',
                         'start_datetime','end_datetime']

class HPOReport():
    def __init__(self, data=None, verbose=True):
        self.verbose = verbose
        self.reset()
        if data:
            self.append(data)

    def reset(self):
        self.attrib = {}
        self.data = []

    def set_primary(self, data):
        self.attrib['project_name'] = data['project_name']
        self.attrib['hyperparameters'] = data['hyperparameters']
        self.attrib['metric'] = data['metric']
        self.attrib['mode'] = data['mode']
        if self.verbose:
            print('INFO: Setting up HPO report with the following attributes')
            print('Project Name     : {project_name} \n'
                  'Hyperparameters  : {hyperparameters}\n'
                  'Metric           : {metric}\n'
                  'Mode             : {mode}'.format(**self.attrib))

    def check_consistency(self, data):
        if not validate_job_metadata(data):
            if self.verbose:
                print('ERROR: Invalid format for HPO result summary. '
                    'Result summary will be skipped')
            return False

        if not self.data:
            self.set_primary(data)

        if (data['project_name'] != self.attrib['project_name']):
            if self.verbose:
                print('INFO: Expecting HPO project name "{}" but "{}" is received. '
                    'Result summary will be skipped.')
            return False
        ####################################################################################### 
        if (set(self.attrib['hyperparameters']) != set(data['hyperparameters'])):
            if self.verbose:
                print('INFO: Inconsistent hyperparameter space. '
                        'Result summary will be skipped.')
            return False
        #######################################################################################
        if (self.attrib['metric'] != data['metric']) or (self.attrib['mode'] != data['mode']):
            if self.verbose:
                print('INFO: Inconsistent metric definition. '
                    'Result summary will be skipped.')
            return False
        return True

    def append(self, data, extras=None):
        if isinstance(data, str):
            data = json.load(open(data))
        if not isinstance(data, list):
            data = [data]
        for d in data:
            if not self.check_consistency(d):
                return None
            keys_to_save = ['start_datetime', 'start_timestamp',
                            'end_datetime', 'task_time_s', 'best_config', 'result']
            result = {key: d[key] for key in keys_to_save}
            if extras:
                for key in extras:
                    result[key] = extras[key]
            self.data.append(result)

    def _check_data(self):
        if not self.data:
            print('ERROR: No HPO results to report.')
            return False
        return True

    @classmethod
    def from_json(cls, fname, verbose=True):
        with open(fname, 'r') as json_file:
            data = json.load(json_file)
        return cls(data, verbose)
    
    @classmethod
    def from_summary(cls, fname, verbose=True):
        with open(fname, 'r') as json_file:
            data = json.load(json_file)
        instance = cls()
        instance.data = data.pop('jobs', None)
        instance.attrib = data
        return instance

    def merge_data(self, extra_columns=None, skip_time=True,
                   metric_min=None, metric_max=None, sort_by_metric=True):
        merged_data = []
        for data in self.data:
            for index in data['result']:
                result = copy.deepcopy(data['result'][index])
                merged_data.append(result)
        if extra_columns:
            extras = []
            for data in self.data: 
                for index in data['result']:
                    extras.append({col: data[col] for col in extra_columns})
            for data, extra in zip(merged_data, extras):
                data.update(extra)
        if skip_time:
            for data in merged_data:
                data.pop('time_s')
        df = pd.DataFrame(merged_data)
        
        if self.attrib['mode'] == 'min':
            sort_ascending = True
        elif self.attrib['mode'] == 'max':
            sort_ascending = False
        
        if sort_by_metric:
            df = df.sort_values(by=self.attrib['metric'] , ascending=sort_ascending).reset_index(drop=True)
        if metric_min is not None:
            df = df[df[self.attrib['metric']] > metric_min]
        if metric_max is not None:
            df = df[df[self.attrib['metric']] < metric_max]
        merged_data = df.to_dict('records')
        return merged_data

    def to_summary(self, fname, indent=2):
        if not self._check_data():
            return None
        report = self.attrib
        report['jobs'] = self.data
        with open(fname, 'w') as file:
            json.dump(report, file, indent=indent)

    def to_json(self, fname, extra_columns=None, skip_time=True, 
                metric_min=None, metric_max=None, indent=2):
        if not self._check_data():
            return None
        data = self.merge_data(extra_columns, skip_time, 
                               metric_min=metric_min,
                               metric_max=metric_max)
        with open(fname, 'w') as output:
            json.dump(data, output, indent=indent)

    def to_dataframe(self, extra_columns=None, skip_time=True, metric_min=None, 
                     metric_max=None, sort_by_metric=True):
        if not self._check_data():
            return None        
        data = self.merge_data(extra_columns, skip_time, 
                               metric_min=metric_min,
                               metric_max=metric_max,
                               sort_by_metric=sort_by_metric)
        df = pd.DataFrame(data)
        return df

    def to_mlflow(self, extra_columns=None, skip_time=True,
                 metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        from mlflow.tracking import MlflowClient
        client = MlflowClient()
        experiment_id = client.create_experiment(self.attrib['project_name'])
        hyperparameters = self.attrib['hyperparameters']
        metric = self.attrib['metric']
        
        if not extra_columns:
            extra_columns = ['start_timestamp']
        elif 'start_timestamp' not in extra_columns:
            extra_columns.append('start_timestamp')
        merged_data = self.merge_data(extra_columns, skip_time=False, 
                                       metric_min=metric_min,
                                       metric_max=metric_max)
        for data in merged_data:
            start_time = int(data['start_timestamp']*1000)
            end_time = int((data['start_timestamp']+data['time_s'])*1000)
            run = client.create_run(experiment_id, start_time=start_time)
            run_id = run.info.run_id
            for hp in hyperparameters:
                client.log_param(run_id, hp, data[hp])
            client.log_metric(run_id, metric, data[metric])
            client.set_terminated(run_id, end_time=end_time)
            
    def to_html(self, fname=None, extra_columns=None, skip_time=True,
               metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        df = self.to_dataframe(extra_columns, skip_time,
                               metric_min=metric_min,
                               metric_max=metric_max)                               
        html_text = df.to_html()
        if fname:
            with open(fname, 'w') as outfile:
                outfile.write(html_text)
        return html_text

    def to_parallel_coordinate_plot(self, fname=None, extra_columns=None, skip_time=True,
                                   metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        import hiplot as hip
        merged_data = self.merge_data(extra_columns, skip_time, 
                                       metric_min=metric_min,
                                       metric_max=metric_max)

        html_text = hip.Experiment.from_iterable(merged_data).to_html()
        if fname:
            with open(fname, 'w') as outfile:
                outfile.write(html_text)
        return html_text

    def to_csv(self, fname, extra_columns=None, skip_time=True,
               metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        df = self.to_dataframe(extra_columns, skip_time, 
                               metric_min=metric_min,
                               metric_max=metric_max)
        csv = df.to_csv(fname, encoding='utf-8')
        return csv

    def show(self, extra_columns=None, skip_time=True, metric_min=None, metric_max=None):
        if not self._check_data():
            return None
        df = self.to_dataframe(extra_columns, skip_time, metric_min=metric_min,
                               metric_max=metric_max)
        print(tabulate(df, showindex=True, headers=df.columns, tablefmt="psql",stralign='center'))


import copy
from typing import Optional, List, Dict

from .base_generator import Generator
from hpogrid.search_space import OptunaSpace

class OptunaGenerator(Generator):
    
    def get_searcher(self, search_space:Dict, metric:str, mode:str, seed:Optional[int]=None, **args):
        import optuna as ot

        search_space = OptunaSpace(search_space).get_search_space()
        
        if mode == 'max':
            direction = "maximize"
        elif mode == 'min':
            direction = "minimize"
        else:
            raise ValueError('mode of evaluation metric can only be "min" or "max"')

        pruner  = ot.pruners.NopPruner()
        storage = ot.storages.InMemoryStorage()
        sampler = ot.samplers.TPESampler(seed=seed)

        study = ot.study.create_study(storage=storage, sampler=sampler,
                                      pruner=pruner, study_name=None,
                                      load_if_exists=True, direction=direction)
        
        self._space = search_space
        
        return study

    def _tell(self, searcher, point:Dict, value):
        import optuna as ot
        trial = ot.trial.create_trial(state=ot.trial.TrialState.COMPLETE,
                                      value=value, params=point,
                                      distributions=self._space,
                                      intermediate_values=None)
        searcher.add_trial(trial)

    def _ask(self, searcher, n_points:int =None):
        points = []
        for _ in range(n_points):
            trial = searcher.ask(fixed_distributions=self._space)
            point = trial.params
            points.append(copy.deepcopy(point))
        return points


from typing import Dict, Tuple

from hpogrid.search_space import NeverGradSpace
from .base_algorithm import BaseAlgorithm

class NeverGradAlgorithm(BaseAlgorithm):
    
    kDefaultBudget = 100
    kDefaultMethod = "RandomSearch"
    
    def __init__(self, metric:str, mode:str, search_space:Dict):
        super().__init__(metric=metric, mode=mode, search_space=search_space)

    def _get_algorithm(self, **args):
        
        search_space = NeverGradSpace(self.search_space).get_search_space()
        
        method = args.pop('method', self.kDefaultMethod)
        import nevergrad as ng
        optimizer = ng.optimizers.registry[method](
                parametrization=search_space, budget=self.kDefaultBudget)
        
        from ray.tune.suggest.nevergrad import NevergradSearch
        algorithm = NevergradSearch(optimizer, None, metric=self.metric, mode=self.mode, **args)
        
        return algorithm
    
    def _restore_search_points(self, algorithm, restore_points:Tuple):
        from hpogrid.generators import NeverGradGenerator
        generator = NeverGradGenerator(search_space=self.search_space,
                                       metric=self.metric, mode=self.mode)
        generator.searcher = algorithm._nevergrad_opt
        
        generator.feed(restore_points[0], restore_points[1])    


from hpogrid.core.defaults import kHPOGridMetadataFormat

def validate_job_metadata(data):
    if not isinstance(data, dict):
        return False
    return all(key in kHPOGridMetadataFormat for key in data)


import os

from ray.tune import Trainable

from hpogrid.core import environment_settings as es
from hpogrid.utils.helper import cd

class CustomModel(Trainable):
    
    def initialize(self, config):
        raise NotImplementedError('the "initialize" method must be overriden for initializing the train model.')
        
    def setup(self, config):
        workdir = es.get_workdir()
        with cd(workdir) as _:
            self.initialize(config)


import os
import sys
import glob
import json
import fnmatch
from typing import Dict, List, Union, Optional

from .defaults import (kConfigPaths, kProjectConfigNameJson, RunMode, ConfigType)
from .environment_settings import get_workdir, get_run_mode

def get_base_path():
    import hpogrid
    return hpogrid.userdata_path        

def get_config_dir(config_type:Union[str, ConfigType]):
    """Returns the directory that stores configuration files of a particular configuration type
    
    Args:
        config_type: str
            Type of configuration
    """
    config_type = ConfigType.parse(config_type)
    config_dir = os.path.join(get_base_path(), kConfigPaths[config_type])
    return config_dir
    
def get_project_path(proj_name:str):
    if get_run_mode() in [RunMode.GRID, RunMode.IDDS]:
        proj_path = get_workdir()
    elif get_run_mode() == RunMode.LOCAL:
        proj_dir = get_config_dir(ConfigType.PROJECT)
        proj_path = os.path.join(proj_dir, proj_name)
    return proj_path

def get_project_script_dir(proj_name:str):
    project_path = get_project_path(proj_name)
    project_script_path = os.path.join(project_path, "scripts")
    return project_script_path

def get_project_config_dir(proj_name:str):
    project_path = get_project_path(proj_name)
    project_config_path = os.path.join(project_path, "config")
    return project_config_path

def get_project_config_path(proj_name:str):
    project_config_dir = get_project_config_dir(proj_name)
    config_path = os.path.join(project_config_dir, kProjectConfigNameJson)
    return config_path
                        
def get_project_config(proj_name:str):
    config_path = get_project_config_path(proj_name)
    if not os.path.exists(config_path):
        raise FileNotFoundError(f'Missing project configuration file: {config_path}')
    config = load_configuration(config_path)
    return config

def get_config_list(config_type:Union[str, ConfigType], expr=None) -> List[str]:
    config_type = ConfigType.parse(config_type)
    if config_type == ConfigType.PROJECT:
        # takes into account when running jobs externally
        if get_run_mode() in [RunMode.GRID, RunMode.IDDS]:
            config_path = get_project_config_path(None)
            config = load_configuration(config_path)
            config_list = [config['project_name']]
        elif get_run_mode() == RunMode.LOCAL:
            config_dir = get_config_dir(config_type)
            config_list = [d for d in os.listdir(config_dir) if 
                           os.path.isdir(os.path.join(config_dir, d))]
    else:
        config_dir = get_config_dir(config_type)
        config_list = glob.glob(os.path.join(config_dir, "*"))
        config_list = [os.path.basename(p) for p in config_list]
        config_list = [os.path.splitext(p)[0] for p in config_list]
    # filter config name by wildcard
    expr = expr if expr is not None else '*'
    config_list = fnmatch.filter(config_list, expr)

    return config_list

def get_project_list(expr=None):
    return get_config_list(ConfigType.PROJECT, expr=expr)

def is_project(obj:str):
    if not isinstance(obj, str):
        return False
    return obj in get_project_list()

def load_configuration(config_input:[Dict, str]) -> Dict:
    """Returns project configuration given input
    
    Args:
        config_input: dict, str
            If dict, it is the configuration itself
            If str, it is either the name of the project,
            of path to a configuration file.
    """
    if isinstance(config_input, dict):
        return config_input
    elif isinstance(config_input, str):
        ext = os.path.splitext(config_input)[1]
        # check if config_input is a project name
        if not ext:
            if is_project(config_input):
                return get_project_config(config_input)
            else:
                raise ValueError(f'project "{config_input}" does not exist')
        # check if configuration file exists
        if not (os.path.exists(config_input)):
            raise FileNotFoundError(f"configuration file {config_input} does not exist.")
        # load configuration file
        with open(config_input, 'r') as file:
            if ext in ['.txt', '.yaml']:
                config = yaml.safe_load(file)
            elif ext == '.json':
                config = json.load(file)
            else:
                raise ValueErrror(f'the configuration file has an unsupported '
                                  f'file extension: {txt} (allowed: txt, yaml, json)')
        return config
    else:
        raise ValueError(f'unknown configuration format: {config_input}')


import os
import json
import yaml
import shutil
from copy import deepcopy
from datetime import datetime
from typing import Dict, List, Union, Optional

from hpogrid import semistaticmethod
from hpogrid.core.defaults import kProjectConfigNameJson, ConfigType, ConfigAction
from hpogrid.core.config_manager import (get_project_script_dir, get_project_config_dir,
                                         get_project_config_path, get_project_path)
from hpogrid.utils.helper import copytree
from hpogrid.configuration import (ConfigurationBase, ModelConfiguration, 
                                   HPOConfiguration, GridConfiguration,
                                   SearchSpaceConfiguration)

class ProjectConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = ConfigType.PROJECT
    _CONFIG_DISPLAY_NAME_ = 'project'
    _LIST_COLUMNS_ = ['Project Title']
    _OUTPUT_EXPR_ = get_project_config_path("{name}")
    _CONFIG_FORMAT_ = \
        {
            'scripts_path': {
                'abbr': 'p',
                'description': 'Path to the location of training scripts'+\
                               ' (or the directory containing the training scripts)',
                'required': True,
                'type': str
            },
            'model_config': {
                'abbr': 'm',
                'description': 'Name of the model configuration to use',
                'required': True,
                'type': str
            },
            'search_space': {
                'abbr': 's',
                'description': 'Name of the search space configuration to use',
                'required': True,
                'type':str
            },
            'hpo_config': {
                'abbr': 'o',
                'description': 'Name of the hpo configuration to use',
                'required': True,
                'type':str
            },
            'grid_config': {
                'abbr': 'g',
                'description': 'Name of the grid configuration to use',
                'required': True,
                'type':str
            }
        }
    
    _CONFIG_CLS_ = {
        "model_config" : ModelConfiguration,
        "hpo_config"   : HPOConfiguration,
        "grid_config"  : GridConfiguration,
        "search_space" : SearchSpaceConfiguration
    }
    
    def __init__(self, name:Optional[str]=None, verbosity:str="INFO"):
        super().__init__(name=name, verbosity=verbosity)
        self._scripts_path = None
    
    def _validate_arguments(self, **args):
        args = super()._validate_arguments(**args)
        scripts_path = args.pop('scripts_path', None)
        if (scripts_path is None):
            self.stdout.info('INFO: Path to training scripts is not specified. Skipping...')
        elif (scripts_path is not None) and (not os.path.exists(scripts_path)):
            raise FileNotFoundError(f'Path to training scripts {scripts_path} does not exist.')
            
        self._scripts_path = scripts_path

        # check if input configuration files exist
        for config_type, config_cls in self._CONFIG_CLS_.items():
            if (config_type in args) and (args[config_type] is not None):
                config_name = args[config_type]
                args[config_type] = config_cls.load_config(config_name)
            else:
                self.stdout.info('INFO: Path to {} configuration is not specified. '
                                 'Skipping...'.format(config_cls._CONFIG_DISPLAY_NAME_))
        return args
    
    @classmethod
    def validate(cls, config:Dict):
        validated_config = deepcopy(config)
        for config_type, config_cls in cls._CONFIG_CLS_.items():
            validated_config[config_type] = config_cls.validate(config[config_type])
        return validated_config
    
    @classmethod
    def _validate(cls, config:Dict):
        cls.stdout.info('INFO: Validating project configuration...')
        result = cls.validate(config)
        cls.stdout.info('INFO: Validation complete.')
        return result
    
    @semistaticmethod
    def _save_scripts(self, source:str, dest:str):
        if os.path.exists(source):
            # copy training scripts to the project directory
            self.stdout.info(f'INFO: Copying training scripts from {source} to {dest}')
            # copy contents of directory to project/scrsipts/
            if os.path.isdir(source):
                copytree(source, dest)
            else:
                shutil.copy2(source, dest)
        else:
            raise FileNotFoundError(f'Path to training scripts {source} does not exist.')
    
    def move_to_backup(self, project_path:str):
        if not os.path.exists(project_path):
            raise FileNotFoundError(f"project path \"{project_path}\" does not exist")
        backup_dir = get_project_path("backup")
        if not os.path.exists(backup_dir):
            os.makedirs(backup_dir)
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        name = os.path.basename(project_path)
        backup_project_path = os.path.join(backup_dir, f'{name}_{timestamp}')
        shutil.move(project_path, backup_project_path)
        return backup_project_path
    
    def save(self, name=None, config=None, scripts_path=None, 
             action:Union[ConfigAction, str]='create'):
        if (name is None) and (config is None):
            name = self._name
            config = self._config
        if name is None: 
            raise ValueError('configuration name undefined')
        config = {'project_name':name, **config}
        
        project_dir = get_project_path(name)
        
        action = ConfigAction.parse(action)
        action_str = ConfigAction.to_str(action, tense="cont").title()
        print(f'INFO: {action_str} project "{name}"')
        if (os.path.exists(project_dir)):
            if  action == ConfigAction.CREATE:
                self.stdout.error(f'ERROR: Project "{name}" already exists. If you want to overwrite,'
                                   ' use "recreate" or "update" action instead of "create".')
                return None
            elif action == ConfigAction.RECREATE:
                backup_project_path = self.move_to_backup(project_dir)
                self.stdout.info('INFO: Recreating project. Original project moved to backup '
                                 f'directory {backup_project_path}.')
        
        # create directories if not already exist
        scripts_dir = get_project_script_dir(name)
        config_dir  = get_project_config_dir(name)
        if not os.path.exists(project_dir):
            os.makedirs(project_dir)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        if not os.path.exists(scripts_dir):
            os.makedirs(scripts_dir)
        
        if scripts_path is None:
            scripts_path = self._scripts_path
        
        # copy training scripts
        if (scripts_path is not None):
            self._save_scripts(scripts_path, scripts_dir)
        
        config_path_json = self.get_config_path(name, extension='json')
        config_path_yaml = self.get_config_path(name, extension='yaml')

        action_str = ConfigAction.to_str(action, tense="past").title()
        with open(config_path_json, 'w') as config_file_json:
            json.dump(config, config_file_json, indent=2)
            self.stdout.info(f'INFO: {action_str} {self._CONFIG_TYPE_} configuration {config_path_json}')
        with open(config_path_yaml, 'w') as config_file_yaml:
            yaml.dump(config, config_file_yaml, default_flow_style=False, sort_keys=False)
            self.stdout.info(f'INFO: {action_str} {self._CONFIG_TYPE_} configuration {config_path_yaml}')
        self.show(name)
       
    @staticmethod
    def create_project(config:Dict, scripts_path:str=None, action:str='create'):
        project_name = config['project_name']
        project_config = ProjectConfiguration()
        project_config.save(project_name, config, scripts_path=scripts_path, action=action)


import copy
from typing import List, Dict

from .base_generator import Generator
from hpogrid.search_space.nevergrad_space import NeverGradSpace

class NeverGradGenerator(Generator):
    
    kDefaultBudget = 100
    kDefaultMethod = "RandomSearch"
    
    def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
        search_space = NeverGradSpace(search_space).get_search_space()
        if ('method' in args) and (args['method']):
            method = args['method']
        else:
            method = self.kDefaultMethod
        import nevergrad as ng
        searcher = ng.optimizers.registry[method](
                parametrization=search_space, budget=self.kDefaultBudget)
        return searcher

    def _ask(self, searcher, n_points:int = None):
        points = []
        for _ in range(n_points):
            point = searcher.ask().kwargs
            points.append(copy.deepcopy(point))
        return points

    def _tell(self, searcher, point:Dict, value):
        value = self._to_metric_values(value)
        searcher.suggest(**point)
        candidate = searcher.ask()
        searcher.tell(candidate, self.signature * value)



import os
import sys
import json
import shutil
import numbers
import inspect
import tarfile
from typing import List, Dict
from contextlib import contextmanager
from subprocess import Popen, PIPE
import shlex

import numpy as np

def old_div(a, b):
    """
    DEPRECATED: import ``old_div`` from ``past.utils`` instead.

    Equivalent to ``a / b`` on Python 2 without ``from __future__ import
    division``.

    TODO: generalize this to other objects (like arrays etc.)
    """
    if isinstance(a, numbers.Integral) and isinstance(b, numbers.Integral):
        return a // b
    else:
        return a / b

def run_command(cmd):
    args = shlex.split(cmd, posix=False)
    with Popen(args, stdout=PIPE, bufsize=1, universal_newlines=True) as p:
        while True:
            line = p.stdout.readline()
            if not line:
                break
            print(line)    
        exit_code = p.poll()
    return exit_code


class NpEncoder(json.JSONEncoder):
    import numpy as np
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)            
        else:
            return super(NpEncoder, self).default(obj)

def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)
            
def pretty_path(path:str):
    return os.path.abspath(os.path.realpath(os.path.expandvars(os.path.expanduser(path))))

def pretty_dirname(path:str):
    path = pretty_path(path)
    if os.path.isdir(path):
        return path
    else:
        return os.path.dirname(path)
    
def is_nan_or_inf(value):
    return np.isnan(value) or np.isinf(value)    
        
@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)

def get_physical_devices(device_type='GPU'):
    import tensorflow as tf
    physical_devices = tf.config.list_physical_devices(device_type)
    return physical_devices

def get_n_gpu():
    #return len(get_physical_devices(device_type='GPU'))
    import torch
    return torch.cuda.device_count()

def get_n_cpu():
    return os.cpu_count()


def extract_tarball(input_path:str, output_path:str) ->List[str]:
    import hpogrid
    tarfiles = [ f for f in os.listdir(input_path) if f.endswith('tar.gz')]
    extracted_files = []
    for f in tarfiles:
        tar = tarfile.open(f, "r:gz")
        hpogrid.stdout.info('INFO: Untaring the file {}'.format(f))
        tar.extractall(path=output_path)
        extracted_files += tar.getnames()
        tar.close()
    return extracted_files

def remove_files(files:List[str]):
    for f in files:
        if os.path.isfile(f):
            os.remove(f)
        elif os.path.isdir(f):
            shutil.rmtree(f)        
            
def is_function_or_method(obj):
    """Check if an object is a function or method.
    Args:
        obj: The Python object in question.
    Returns:
        True if the object is an function or method.
    """
    return inspect.isfunction(obj) or inspect.ismethod(obj) or is_cython(obj)


def is_class_method(f):
    """Returns whether the given method is a class_method."""
    return hasattr(f, "__self__") and f.__self__ is not None  


"""
import socket
hostname = socket.gethostname()
"""



import math
import numpy as np

from .base_space import BaseSpace

from hpogrid.utils.helper import old_div

class PBTSpace(BaseSpace):

    def __init__(self, search_space = None):
        self.library = 'pbt'
        super().__init__(search_space)    
        
    def reset_space(self):
        self.search_space = {}

    def append(self, space_value):
        self.search_space[space_value[0]] = space_value[1]

    def categorical(self, label, categories, grid_search = False):
        return (label, categories)

    def uniform(self, label, low, high):
        return (label, lambda: np.random.uniform(low, high))
    
    def uniformint(self, label, low, high):
        return (label, lambda : np.random.randint(low, high))

    def loguniform(self, label, low, high, base=10):
        logmin = math.log(low) / math.log(base)
        logmax = math.log(high) / math.log(base)
        return (label, lambda: base**(np.random.uniform(logmin, logmax)))

    def qloguniform(self, label, low, high, base=10):
        logmin = math.log(low) / math.log(base)
        logmax = math.log(high) / math.log(base)
        return (label, lambda: 
            round(old_div(base**(np.random.uniform(logmin, logmax)),q))*q)

    def quniform(self, label, low, high, q):
        return (labelm, lambda: 
            round(old_div(np.random.uniform(low, high),q))*q)

    def quniformint(self, label, low, high, q):
        return (label, lambda : 
            round(old_div(np.random.uniform(low, high),q))*q)

    def normal(self, mu, sigma):
        return (label, lambda : sigma*np.random.randn()+mu)

    def qnormal(self, mu, sigma, q):
        return (label, lambda : 
            round(old_div(sigma*np.random.randn()+mu,q))*q)

    def lognormal(self, mu, sigma, base=math.e):
        return (label, lambda : 
            math.pow(sigma*np.random.randn()+mu, base))

    def qlognormal(self, mu, sigma, q, base=math.e):
        return (label, lambda _: 
            round(old_div(math.pow(sigma*np.random.randn()+mu, base),q))*q )    

    def fixed(self, value):
        return (label, [value])


import os
import json
from typing import Optional, Dict, List, Union

import hpogrid
from hpogrid import stdout
from hpogrid.core.defaults import (kDefaultiDDSContainer, kiDDSHPinput, kiDDSHPoutput,
                                   kiDDSBasePath, kDefaultProxyPath)
from hpogrid.core import environment_settings
from hpogrid.utils import helper, stylus

def transpose_config(config:Dict):
    project_name = config['project_name']
    evaluation_container = config['grid_config']['container']
    search_space = config['search_space']
    library = config['hpo_config']['algorithm']
    if config['hpo_config']['algorithm_param']:
        method = config['hpo_config']['algorithm_param'].get('method', None)
    else:
        method = None
    max_point = config['hpo_config']['num_trials']
    max_concurrent = config['hpo_config']['max_concurrent']


    # format generator command
    generator_options = {}
    generator_options['n_point'] = '%NUM_POINTS'
    generator_options['max_point'] = '%MAX_POINTS'
    generator_options['infile'] = os.path.join(kiDDSBasePath, '%IN')
    generator_options['outfile'] = os.path.join(kiDDSBasePath, '%OUT')
    generator_options['lib'] = library
    if (library == 'nevergrad') and method:
        generator_options['method'] = method
    generator_cmd = 'hpogrid generate {}'.format(stylus.join_options(generator_options))

    # format idds configuration
    idds_config = {}
    idds_config['steeringExec'] = ("run --rm -v \"$(pwd)\":{idds_base_path} {container} " +\
    "/bin/bash -c \"{cmd}\" ").format(idds_base_path=kiDDSBasePath, container=kDefaultiDDSContainer, cmd=generator_cmd)
    #idds_config['steeringExec'] = ("run --rm -v \"$(pwd)\":{idds_base_path} {container} " +\
    #"/bin/bash -c \"{cmd}\" ").format(idds_base_path="/HPOiDDS", container="gitlab-registry.cern.ch/zhangruihpc/steeringcontainer:0.0.2", cmd=generator_cmd)

    idds_config['evaluationExec'] = "pip install --upgrade hpogrid && " + \
                                    "hpogrid run {project_name} --mode idds".format(project_name=project_name)
    idds_config['evaluationContainer'] = evaluation_container
    idds_config['evaluationInput'] = kiDDSHPinput
    idds_config['evaluationOutput'] = kiDDSHPoutput
    idds_config['evaluationMeta'] = kGridSiteMetadataFileName
    idds_config['nParallelEvaluation'] = max_concurrent
    idds_config['maxPoints'] = max_point
    return idds_config, search_space


def get_search_points():
    if not environment_settings.is_idds_job():
        raise RuntimeError('Cannot get search points for non-idds environment')
    workdir = helper.get_workdir()
    search_point_path = os.path.join(workdir, kiDDSHPinput)
    if not os.path.exists(search_point_path):
        raise FileNotFoundError('No such file or directory: \'{}\''.format(search_point_path))
    with open(search_point_path,'r') as search_point_file:
        search_points = json.load(search_point_file)
    return search_points

def create_idds_output_from_metadata(metadata:Dict):
    idds_output = {}
    idds_output['status'] = 0
    loss = []
    metric = metadata['metric']
    for index in metadata['result']:
        loss.append(metadata['result'][index][metric])
    if len(loss) == 1:
        loss = loss[0]
    idds_output['loss'] = loss
    idds_output['message'] = ''
    return idds_output


def save_idds_output_from_metadata(metadata):
    if not environment_settings.is_idds_job():
        raise RuntimeError('save_output should not be called in non-idds environment')    
        
    output = create_idds_output_from_metadata(metadata)
    workdir = helper.get_workdir()
    output_path = os.path.join(workdir, kiDDSHPoutput)
    with open(output_path, 'w') as output_file:
        stdout.info('INFO: Saved idds output at {}'.format(output_path))
        json.dump(output, output_file, cls=helper.NpEncoder)
        
        
def set_config_path(path:Optional[str]=None, override:bool=True):
    # use default idds config if not given
    if path is None:
        path = os.path.join(hpogrid.resource_path, "idds.cfg")
    if ("IDDS_CONFIG" not in os.environ) or override:
        os.environ["IDDS_CONFIG"] = path
        
def download_log(self, workload_id:int, request_id=None,
                 idds_config_path=None, proxy_path=kDefaultProxyPath):
    set_config_path(idds_config_path, override=False)
    os.environ['IDDS_CONFIG'] = idds_config_path
    original_proxy = os.environ.get('X509_USER_PROXY', None)
    os.environ['X509_USER_PROXY'] = proxy_path
    from idds.client.client import Client
    from idds.common.utils import get_rest_host
    host = get_rest_host()
    client = Client(host=host)

    stdout.info('INFO: Fetching iDDS log files for the task with taskID = {}'.format(workload_id))
    filename = client.download_logs(workload_id=workload_id, request_id=request_id, dest_dir='/tmp')
    if not filename:
        stdout.info('INFO: No iDDS log found for the task with taskID = {}'.format(workload_id))
    else:
        stdout.info('INFO: iDDS log files can be found at: {}'.format(filename))
    if original_proxy:
        os.environ['X509_USER_PROXY'] = original_proxy
        


from .base_generator import Generator
from .grid_generator import GridGenerator
from .bohb_generator import BOHBGenerator
from .hyperopt_generator import HyperOptGenerator
from .nevergrad_generator import NeverGradGenerator
from .skopt_generator import SkOptGenerator
from .ax_generator import AxGenerator
from .optuna_generator import OptunaGenerator


from .base_algorithm import BaseAlgorithm
from .ax_algorithm import AxAlgorithm
from .bohb_algorithm import BOHBAlgorithm
from .hyperopt_algorithm import HyperOptAlgorithm
from .nevergrad_algorithm import NeverGradAlgorithm
from .skopt_algorithm import SkOptAlgorithm
from .optuna_algorithm import OptunaAlgorithm


import os
import sys
import glob
import json
import tempfile
from typing import Optional, Dict, List

from hpogrid import semistaticmethod
from hpogrid.core.defaults import *
from hpogrid.core import config_manager
from hpogrid.utils import helper, stylus

from .abstract_object import AbstractObject

class GridHandler(AbstractObject):
    
    def __init__(self, verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)
    
    @semistaticmethod
    def submit(self, config_input:str, n_jobs:int=1, site:str=None, time:int=-1, search_points:List=None,
               mode:str='grid', debug:bool=False, scripts_path:str=None, extra:Optional[Dict]=None):
        if extra is None:
            extra = []
        config = self.preprocess_config(config_input, scripts_path)
        if mode == 'grid':
            self.submit_grid(config, n_jobs=n_jobs, site=site, time=time,
                             search_points=search_points, debug=debug, extra=extra)
        elif mode == 'idds':
            self.submit_idds(config, n_jobs=n_jobs, site=site, debug=debug, extra=extra)
        else:
            raise ValueError('Unknown job submission mode: {}'.format(mode))
            
    @semistaticmethod
    def preprocess_config(self, config_input:str, scripts_path:str=None):
        config = config_manager.load_configuration(config_input)
        from hpogrid.configuration import ProjectConfiguration
        config = ProjectConfiguration._validate(config)
        project_name = config['project_name']
        if not config_manager.is_project(project_name):
            self.stdout.info(f'INFO: Creating project directory for "{project_name}"')
            ProjectConfiguration.create_project(config, scripts_path=scripts_path, action='create')
        elif not config_manager.is_project(config_input):
            self.stdout.info(f'INFO: Replacing project directory for "{project_name}"')
            ProjectConfiguration.create_project(config, scripts_path=scripts_path, action='recreate')
        return config
        
    @semistaticmethod
    def submit_idds(self, config:Dict, n_jobs:int=1, site:str=None, debug:bool=False,
                    extra:Optional[Dict]=None):
        if extra is None:
            extra = []
        project_path = config_manager.get_project_path(config['project_name'])
        self.stdout.info(f'INFO: Submitting {n_jobs} idds grid job(s)')
        with tempfile.TemporaryDirectory() as tmpdirname:
            from hpogrid.interface.idds.main import transpose_config
            idds_config, search_space = transpose_config(config)
            idds_config_path = os.path.join(tmpdirname, 'idds_config.json')
            search_space_path = os.path.join(tmpdirname, 'search_space.json')
            # create temporary files for idds configuration and search space
            with open(idds_config_path, 'w') as idds_config_file:
                json.dump(idds_config, idds_config_file)
            with open(search_space_path, 'w') as search_space_file:
                json.dump(search_space, search_space_file)  
            command = self.format_idds_command(config, tmpdirname, site=site, extra=extra)
            with helper.cd(project_path):
                if debug:
                    self.stdout.info('INFO: iDDS configuration is')
                    table = stylus.create_formatted_dict(idds_config, ['Attribute', 'Value'], indexed=False)
                    self.stdout.info(table)
                    # submit jobs
                    self.stdout.info('INFO: The idds command is ')
                    self.stdout.info(f'{command}')
                for _ in range(n_jobs):
                    os.system(command)
    @semistaticmethod
    def submit_grid(self, config:Dict, n_jobs:int=1, site:str=None, time:int=-1, debug:bool=False,
                    search_points:List=None, scripts_path:str=None, extra:Optional[Dict]=None):
        if extra is None:
            extra = []
        project_path = config_manager.get_project_path(config['project_name'])
        print(f'INFO: Submitting {n_jobs} grid job(s)')
        command = self.format_grid_command(config, site=site, time=time,
                                           search_points=search_points,  extra=extra)
        with helper.cd(project_path):
            # submit jobs
            for _ in range(n_jobs):
                if debug:
                    self.stdout.info('INFO: The prun command is ')
                    self.stdout.info(f'{command}')
                os.system(command)
                
    @staticmethod
    def format_idds_command(config:Dict, tmpdirname:str, site:str=None, extra:Optional[Dict]=None):
        if extra is None:
            extra = []
        idds_config_path = os.path.join(tmpdirname, 'idds_config.json')
        search_space_path = os.path.join(tmpdirname, 'search_space.json')
        options = {'loadJson': idds_config_path,
                   'searchSpaceFile': search_space_path}
        proj_name = config['project_name']
        grid_config = config['grid_config']
        # override site settings if given
        if not site:
            site = grid_config['site']
                
        if site:
            if isinstance(site, str):
                if site != 'ANY':
                    options['site'] = site
            elif isinstance(site, list):
                options['site'] = ','.join(site)
            else:
                raise ValueError('Invalid site settings: {}'.format(site))
            # specify architecture for gpu/cpu site
            if 'GPU' in options['site']:
                options['architecture'] = 'nvidia-gpu'
        # options excluded from idds configuration file due to
        # possible bash variable expansion
        in_ds = config['grid_config']['inDS']
        out_ds = config['grid_config']['outDS']
        
        if in_ds:
            options['trainingDS'] = in_ds
            
        if '{HPO_PROJECT_NAME}' in out_ds:
            out_ds = out_ds.format(HPO_PROJECT_NAME=proj_name)
        options['outDS'] = out_ds
                
        command = 'phpo {} {}'.format(stylus.join_options(options), ' '.join(extra))    
        return command
    
    @staticmethod
    def format_grid_command(config:Dict, site:str=None, time:int=-1,
                            search_points:List=None, extra:Optional[Dict]=None):
        if extra is None:
            extra = []
        options = {'forceStaged': '',
                   'useSandbox': '',
                   'noBuild': '',
                   'alrb': ''}
        
        grid_config = config['grid_config']
        proj_name = config['project_name']
        
        options['containerImage'] = grid_config['container']
        
        search_points = '' if not search_points else '--search_points {}'.format(search_points)

        options['exec'] = '"pip install --upgrade hpogrid && '+\
                          'hpogrid run {} --mode grid {}"'.format(proj_name, search_points)

        if not grid_config['retry']:
            options['disableAutoRetry'] = ''

        if grid_config['inDS']:
            options['inDS'] = grid_config['inDS']

        if '{HPO_PROJECT_NAME}' in grid_config['outDS']:
            grid_config['outDS'] = grid_config['outDS'].format(HPO_PROJECT_NAME=proj_name)
                
        options['outDS'] = grid_config['outDS']
                
        if time != -1:
            options['maxCpuCount'] = str(time)
        
        # override site settings if given
        if not site:
            site = grid_config['site']
                
        if site:
            if isinstance(site, str):
                if site != 'ANY':
                    options['site'] = site
            elif isinstance(site, list):
                options['site'] = ','.join(site)
            else:
                raise ValueError('Invalid site settings: {}'.format(site))
            # specify architecture for gpu/cpu site
            if 'GPU' in options['site']:
                options['architecture'] = 'nvidia-gpu'
        
        command = 'prun {} {}'.format(stylus.join_options(options), ' '.join(extra))
        return command        
                


import os
from typing import Optional, List, Dict

from hpogrid.components import HPOReport

class PandaTaskReport(HPOReport):
    
    def __init__(self):
        super().__init__()
        from hpogrid.components.panda_task_manager import PandaTaskManager
        self.taskmgr = PandaTaskManager()
        
    def display(self, project_name, outname="hpo_result", metric_min=None,
                metric_max=None, extra_columns=None, taskid_range=None, 
                summary=False, to_json=False, to_html=False,
                to_csv=False, to_pcp=False, to_mlflow=False,
                **task_params):
        task_params['metadata'] = True
        task_params['status'] = ['finished', 'done']
        datasets = self.taskmgr.query_tasks(**task_params)
        # filter tasks by jetitaskid range
        if taskid_range is not None:
            datasets = self.taskmgr.filter_range(datasets, taskid_range)
        col = ['computingsite', 'jobs_metadata', 'jeditaskid', 'site']
        datasets = self.taskmgr.filter_datasets(datasets, col)
        
        extra_columns = extra_columns if extra_columns is not None else []
        for extra in extra_columns:
            if not extra in kSupportedExtraColumns:
                raise ValueError('Unsupported extra column {}'.format(extra))

        skip_time = 'tims_s' not in extra_columns
        self.parse_datasets(datasets, project_name, extra_columns=extra_columns)
        self.show(extra_columns, skip_time)

        bounds = {'metric_min':metric_min, 'metric_max':metric_max}
        
        if summary:
            self.to_summary(outname+'_summary.json')
        if to_json:
            self.to_json(outname+'.json', extra_columns, skip_time, **bounds)
        if to_html:
            self.to_html(outname+'.html', extra_columns, skip_time, **bounds)
        if to_csv:
            self.to_csv(outname+'.csv', extra_columns, skip_time, **bounds)    
        if to_pcp:
            self.to_parallel_coordinate_plot(outname+'_parallel_coordinate_plot.html',
                                             extra_columns, skip_time, **bounds)   
        if to_mlflow:
            self.to_mlflow(extra_columns, skip_time, **bounds)

    def filter_invalid_datasets(self, datasets, project_name):
        filtered_datasets = []
        metadata_list = []
        for dataset in datasets:
            metadata = self.extract_metadata(dataset)
            # check if metadata is empty
            if not metadata:
                continue
            # check if it is a valid metadata generated by hpogrid
            # and if the metadata corresponds to the project of concern
            if not all( validate_job_metadata(data) and 
                       data.get('project_name', None) == project_name \
                       for data in metadata):
                continue
            filtered_datasets.append(dataset)
            metadata_list += metadata
        if len(filtered_datasets) == 0:
            print('INFO: No results found.')
            sys.exit(1)
        metrics = set([metadata['metric'] for metadata in metadata_list])
        if not len(metrics)==1 :
            raise RuntimeError('HPO results from different tasks consist of different metrics: {}'.format(metrics))
        hparams = [metadata['hyperparameters'] for metadata in metadata_list]
        if not all(hp == hparams[0] for hp in hparams):
            raise RuntimeError('HPO results from different tasks consist of different hyperparameter space')
        return filtered_datasets

    def extract_metadata(self, dataset):
        metadata_list = []
        if 'jobs_metadata' in dataset:
            jobs_metadata = dataset['jobs_metadata']
            # take care of general case where a job contains multiple sub-jobs
            # which will return a list of metadata
            if not isinstance(jobs_metadata, list):
                jobs_metadata = [jobs_metadata]
            for job_metadata in jobs_metadata:
                if not isinstance(job_metadata, dict):
                    raise ValueError('Invalid job metadata format')
                keys = list(job_metadata.keys())
                # skip empty metadata
                if not keys:
                    continue
                metadata = job_metadata[keys[0]].get('user_job_metadata', None)
                if metadata is not None:
                    metadata_list.append(metadata)
                else:
                    print('WARNING: The key "user_job_metadata" not found in job metadata. '
                          'Probably the format has changed. Skipping.')
        return metadata_list

    def parse_datasets(self, datasets, project_name, extra_columns=kDefaultExtraColumns):
        summary = []
        valid_datasets = self.filter_invalid_datasets(datasets, project_name)

        metric =  None
        mode = None
        
        for dataset in valid_datasets:
            metadata = self.extract_metadata(dataset)
            extras = {}
            if 'taskid' in extra_columns:
                extras['taskid'] = dataset['jeditaskid']
            if 'site' in extra_columns:
                if 'computingsite' in dataset:
                    extras['site'] = dataset['computingsite']
                elif 'site' in dataset:
                    extras['site'] = dataset['site']
                else:
                    extras['site'] = ''
            self.append(metadata, extras)


from .base_space import BaseSpace
from .ax_space import AxSpace
from .bohb_space import BOHBSpace
from .hyperopt_space import HyperOptSpace
from .nevergrad_space import NeverGradSpace
from .pbt_space import PBTSpace
from .skopt_space import SkOptSpace
from .tune_space import TuneSpace
from .optuna_space import OptunaSpace


from hpogrid.core.defaults import ConfigType
from hpogrid.configuration import ConfigurationBase

class ModelConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = ConfigType.MODEL
    _CONFIG_DISPLAY_NAME_ = 'model'
    _LIST_COLUMNS_ = ['Model Configuration']


"""
    Module for formatting tables, texts and figures used in hpogrid
"""
import yaml
import copy

import pandas as pd
from tabulate import tabulate

kDefaultTableStyle = 'psql'
kDefaultStrAlign = 'left'

class ColorCode():
    RED = '\033[0;91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    MAGENTA = '\033[95m'
    CYAN = '\033[96m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    BLINK = '\033[5m'
    
def str_presenter(dumper, data):
    """configures yaml for dumping multiline strings
    Ref: https://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data"""
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)

yaml.add_representer(str, str_presenter)
yaml.representer.SafeRepresenter.add_representer(str, str_presenter)    
    
def insert_newlines(string, every=64):
    return '\n'.join(string[i:i+every] for i in range(0, len(string), every))

def type2str(typevar):
    if isinstance(typevar, tuple):
        return ', '.join([t.__name__ for t in typevar])
    else:
        return typevar.__name__    

def join_options(options):
    return ' '.join(["--{} {}".format(key,value) for (key,value) in options.items()])

def create_table(data, columns=None, indexed=True, transpose=False,
    tableformat=kDefaultTableStyle, stralign=kDefaultStrAlign):
    df = pd.DataFrame(data, columns=columns)
    if transpose:
        df = df.transpose()
    table = tabulate(df, showindex=indexed, headers=df.columns, 
        tablefmt=tableformat,
        stralign=stralign)
    return table

def format_dict_str_len(source):
    if isinstance(source, dict):
        for key, value in source.items():
            source[key] = format_dict_str_len(value)
        return source
    elif isinstance(source, str) and (len(source) > 80):
        return insert_newlines(source, 80)
    else:
        return source

def create_formatted_dict(data, columns=None, indexed=True, transpose=False,
    tableformat=kDefaultTableStyle, stralign=kDefaultStrAlign):
    _data = copy.deepcopy(data)
    _data = format_dict_str_len(_data)  
    if isinstance(_data, dict):
        for key in _data:
            if isinstance(_data[key], dict):
                _data[key] = yaml.dump(_data[key], allow_unicode=True,
                                       default_flow_style=False, sort_keys=False)
    df = pd.DataFrame(_data.items(), columns=columns)
    if transpose:
        df = df.transpose()
    
    table = tabulate(df, showindex=indexed, headers=df.columns, 
        tablefmt=tableformat,
        stralign=stralign)
    return table


from typing import Dict, Tuple

from hpogrid.search_space import SkOptSpace
from .base_algorithm import BaseAlgorithm

class SkOptAlgorithm(BaseAlgorithm):
    
    def __init__(self, metric:str, mode:str, search_space:Dict):
        super().__init__(metric=metric, mode=mode, search_space=search_space)
        
    def _get_algorithm(self, **args):
        
        search_space = SkOptSpace(self.search_space).search_space
        
        from skopt import Optimizer
        optimizer = Optimizer(search_space)
        labels = [hp.name for hp in search_space]
        
        from ray.tune.suggest.skopt import SkOptSearch
        algorithm = SkOptSearch(optimizer, labels, metric=self.metric, mode=self.mode, **args)
        
        return algorithm
    
    def _restore_search_points(self, algorithm, restore_points:Tuple):
        from hpogrid.generators import SkOptGenerator
        generator = SkOptGenerator(search_space=self.search_space,
                                   metric=self.metric, mode=self.mode)
        generator.searcher = algorithm._skopt_opt
        
        generator.feed(restore_points[0], restore_points[1])


from typing import Dict, Tuple

from hpogrid.search_space import OptunaSpace
from .base_algorithm import BaseAlgorithm

class OptunaAlgorithm(BaseAlgorithm):
    
    def __init__(self, metric:str, mode:str, search_space:Dict):
        super().__init__(metric=metric, mode=mode, search_space=search_space)
        
    def _get_algorithm(self, **args):
        
        search_space = OptunaSpace(self.search_space).search_space
        
        from ray.tune.suggest.optuna import OptunaSearch
        algorithm = OptunaSearch(space=search_space, metric=self.metric, mode=self.mode, **args)
        
        return algorithm
    
    def _restore_search_points(self, algorithm, restore_points:Tuple):
        from hpogrid.generators import OptunaGenerator
        generator = OptunaGenerator(search_space=self.search_space,
                                    metric=self.metric, mode=self.mode)
        generator.searcher = algorithm._ot_study
        
        generator.feed(restore_points[0], restore_points[1])


import math

from .base_space import BaseSpace

class HyperOptSpace(BaseSpace):
    def __init__(self, search_space=None):
        self.library = 'hyperopt'
        super().__init__(search_space)

    def reset_space(self):
        self.search_space = {}
        
    def append(self, space_value):
        self.search_space[space_value[0]] = space_value[1]

    def categorical(self, label, categories, grid_search=False):
        if grid_search != False:
            raise ValueError(f'{self.library} does not allow grid search')
        from hyperopt import hp
        return (label, hp.choice(label, categories))

    def uniformint(self, label, low, high):
        from hyperopt import hp
        return (label, hp.uniformint(label, low, high))

    def uniform(self, label, low, high):
        from hyperopt import hp
        return (label, hp.uniform(label, low, high))

    def quniform(self, label, low, high, q):
        return (label, hp.uniform(label, low, high, q))

    def loguniform(self, label, low, high, base = math.e):
        if base != math.e:
            raise ValueError(f'{self.library} search space only allows base e for loguniform sampling')
        from hyperopt import hp
        return (label, hp.loguniform(label, math.log(low), math.log(high)))

    def qloguniform(self, label, low, high, q, base=math.e):
        if base != math.e:
            raise ValueError(f'{self.library} search space only allows base e for qloguniform sampling')
        from hyperopt import hp
        return (label, hp.qloguniform(label, math.log(low), math.log(high)), q)

    def normal(self, label, mu, sigma):
        from hyperopt import hp
        return (label, hp.normal(label, mu, sigma))

    def qnormal(self, label, mu, sigma, q):
        from hyperopt import hp
        return (label, hp.normal(label, mu, sigma, q))

    def lognormal(self, label, mu, sigma, base):
        if base != math.e:
            raise ValueError(f'{self.library} search_space only allows base e for lognormal sampling')
        from hyperopt import hp
        return (label, hp.lognormal(label, mu, sigma))

    def qlognormal(self, label, mu, sigma, q, base=math.e):
        if base != math.e:
            raise ValueError(f'{self.library} search_space only allows base e for qlognormal sampling')
        from hyperopt import hp
        return (label, hp.qlognormal(label, mu, sigma, q))

    def fixed(self, value):
        return (label, value)



from hpogrid.core.defaults import ConfigType
from hpogrid.configuration import ConfigurationBase

class HPOConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = ConfigType.HPO
    _CONFIG_DISPLAY_NAME_ = 'HPO'
    _LIST_COLUMNS_ = ['HPO Configuration']


import os
import sys
import json
from typing import Optional, List, Dict

from ray.tune import Callback

class TuneCallback(Callback):
    
    def custom_setup(self, hyperparameters:Optional[List]=None, metrics:Optional[List]=None, cache_dir:Optional[str]=None, **info):
        
        self._hpogrid_hyperparameters = hyperparameters
        self._hpogrid_metrics = metrics
        self._hpogrid_cache_dir = cache_dir
        
    def get_trial_info(self, trial: "Trial"):
        try:
            name = trial.trainable_name
            pid  = trial._get_default_result_or_future()['pid']
        except:
            name = "?????"
            pid  = "?????"
        return {"name": name, "pid": pid}
    
    def get_msg_prefix(self, trial: "Trial", pid:Optional[int]=None):
        trial_info = self.get_trial_info(trial)
        if pid is not None:
            trial_info['pid'] = pid
        msg_prefix = f"({trial_info['name']} pid={trial_info['pid']})"
        return msg_prefix
    
    def format_message(self, text:str, prefix:str, fill_space:bool=False):
        prefix_str = prefix if not fill_space else " "*len(prefix)
        text = f"\033[36m{prefix_str} {text}\033[0m"
        return text
        
    def on_trial_start(self, iteration: int, trials: List["Trial"],
                       trial: "Trial", **info):
        if self._hpogrid_hyperparameters:
            prefix = self.get_msg_prefix(trial)
            sys.stdout.write(self.format_message("Hyperparameters: \n", prefix))
            for name in self._hpogrid_hyperparameters:
                if name in trial.config:
                    value = trial.config[name]
                    sys.stdout.write(self.format_message(f"\t{name} = {value}\n", prefix))

    def on_trial_result(self, iteration, trials, trial, result, **info):

        if self._hpogrid_metrics:
            prefix = self.get_msg_prefix(trial, pid=result.get("pid", None))
            sys.stdout.write(self.format_message("Results: \n", prefix))
            for name in self._hpogrid_metrics:
                if name in result:
                    value = result[name]
                    sys.stdout.write(self.format_message(f"\t{name} = {value}\n", prefix))

        if self._hpogrid_cache_dir:
            cache_filename = f"timestamp_{result['timestamp']}_trial_{result['trial_id']}.json"
            cache_path = os.path.join(self._hpogrid_cache_dir, cache_filename)
            with open(cache_path, "w") as cache_file:
                json.dump(result, cache_file, indent=2)


import json
from copy import deepcopy
from json import JSONDecodeError

from hpogrid.utils.stylus import type2str
from hpogrid.core.defaults import kSearchSpaceFormat
from hpogrid.configuration import ConfigurationBase

class SearchSpaceConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = 'search_space'
    _CONFIG_DISPLAY_NAME_ = 'search space'
    _LIST_COLUMNS_ = ['Search Space Configuration']

    @classmethod
    def validate(cls, config):
        if 'search_space' in config:
            search_space = deepcopy(config['search_space'])
        else:
            search_space = config
        # if attribute type is dict, parse string input as dict
        if isinstance(search_space, str):
            try:
                search_space = json.loads(search_space)
            except JSONDecodeError:
                raise RuntimeError('ERROR: Cannot decode the value of "search_space" as dictionary.'
                    'Please check your input.')        
        config_format = kSearchSpaceFormat
        sampling_methods = list(config_format)
        kHPKeys = set(['method', 'dimension'])
        for hp in search_space:
            if (not isinstance(search_space[hp], dict)) or (set(search_space[hp].keys()) != kHPKeys):
                raise ValueError('Each hyperparameter must be defined by a dictionary containing the keys "method" and "dimension"')
            if not isinstance(search_space[hp]['dimension'], dict):
                raise ValueError('Dimension of a hyperparameter must be a dictionary containing the parameters of its sampling method')
            method = search_space[hp]['method']
            # check if hyperparameter sampling method is supported
            if method not in sampling_methods:
                raise ValueError('Sampling method "{}" is not supported. '
                                 'Supported methods are: {}'.format(method,','.join(sampling_methods)))
            for arg in config_format[method]:
                if arg in search_space[hp]['dimension']:
                    # check if the argument type of the sampling method is correct
                    value_type = config_format[method][arg]['type']
                    if not isinstance(search_space[hp]['dimension'][arg], value_type):
                        raise ValueError('The value of argument "{}" for the sampling method '
                                         '"{}" (for hyperparameter "{}") must be of type {}'.format(
                                         arg, method, hp, type2str(value_type)))
                else:
                    # check if a required argument for the sampling method is missing
                    if (config_format[method][arg]['required']):
                        raise ValueError('Missing argument "{}" for the sampling method "{}" '
                                         'for the hyperparameter "{}"'.format(arg, method, hp))
                    # set the argument to its default value if not specified
                    if ('default' in config_format[method][arg]):
                        cls.stdout.info('Info: The argument "{}" for the sampling method "{}" for'
                                        ' the hyperparameter "{}" will be set to its default value {}'.format(
                                        arg, method, hp, config_format[method][arg]['default']))
                        search_space[hp]['dimension'][arg] = config_format[method][arg]['default']

        return search_space


from copy import deepcopy
from enum import Enum
from typing import Union, Optional

class GeneralEnum(Enum):
    @classmethod
    def parse(cls, expr:str):
        if isinstance(expr, cls):
            return expr
        _expr = expr.strip().upper()
        if _expr in cls.__members__:
            return cls[_expr]
        else:
            options = list(cls.__members__)
            raise RuntimeError(f"invalid option: {expr} (allowed options: {', '.join(options)})")
    @classmethod
    def get_members(cls):
        return [i.lower() for i in cls.__members__]
    

class RunMode(GeneralEnum):
    LOCAL = 0
    GRID  = 1
    IDDS  = 2

            
class ConfigType(GeneralEnum):
    BASE         = 0
    PROJECT      = 1
    HPO          = 2
    SEARCH_SPACE = 3
    MODEL        = 4
    GRID         = 5

class ConfigAction(GeneralEnum):
    NONE     = 0
    LIST     = 1
    SHOW     = 2
    REMOVE   = 3
    CREATE   = 4
    RECREATE = 5
    UPDATE   = 6
    
    @staticmethod
    def to_str(value, tense:str="past"):
        if value == ConfigAction.CREATE:
            if tense == "past":
                return "created"
            elif tense == "cont":
                return "creating"
        elif value == ConfigAction.RECREATE:
            if tense == "past":
                return "recreated"
            elif tense == "cont":
                return "recreating"
        elif value == ConfigAction.UPDATE:
            if tense == "past":
                return "udpated"
            elif tense == "cont":
                return "updating"
        return value

class GeneratorMethod(GeneralEnum):
    GRID      = 0
    SKOPT     = 1
    NEVERGRAD = 2
    HYPEROPT  = 3
    AX        = 4
    BOHB      = 5
    OPTUNA    = 6
    
class SchedulerType(GeneralEnum):
    ASYNCHYPERBAND = 0
    BOHBHYPERBAND  = 1
    PBT            = 2
    
class SearchSpaceType(GeneralEnum):
    AX        = 0
    BOHB      = 1
    HYPEROPT  = 2
    NEVERGRAD = 3
    PBT       = 4
    SKOPT     = 5
    TUNE      = 6
    OPTUNA    = 7
    
class AlgorithmType(GeneralEnum):
    AX        = 0
    BOHB      = 1
    HYPEROPT  = 2
    NEVERGRAD = 3
    SKOPT     = 4
    TUNE      = 5
    RANDOM    = 6
    GRID      = 7
    BAYESIAN  = 8
    OPTUNA    = 9

kProjectConfigNameJson = 'config.json'
kProjectConfigNameYaml = 'config.yaml'

kiDDSConfigName = 'idds_config.json'
kiDDSSearchSpaceName = 'search_space.json'
kiDDSHPinput = 'input.json'
kiDDSHPoutput = 'output.json'
kiDDSMetrics = "metrics.tgz"

#####################################################
##############    Paths Related      ################
#####################################################

# user config paths relative to userdata/
kConfigPaths = {
    ConfigType.PROJECT      : 'projects',
    ConfigType.HPO          : 'hpo',
    ConfigType.SEARCH_SPACE : 'search_space',
    ConfigType.MODEL        : 'model',
    ConfigType.GRID         : 'grid'
}

kDockerBasePath = '/workDir'
kiDDSBasePath = kDockerBasePath
kGridSiteWorkDir = '/srv/workDir/'
kDefaultProxyPath = '/etc/pki/tls/cert.pem'

#####################################################
############    Grid Site Related      ##############
#####################################################

kGPUSiteNGPU = {
    'ANALY_MANC_GPU_TEST': 10, #single queue, no submission parameters, 1 GPU per job
    'ANALY_QMUL_GPU_TEST': 6, # GPUNumber=x for now is hardcoded in the dev APF JDL,number of GPUs per job limited by cgroups, K80=2*K40, so total of 6 gpu slots avalable.
    'ANALY_MWT2_GPU': 8, #single queue, no submission parameters, 1 GPU per job
    'ANALY_BNL_GPU_ARC': 12, #also shared with Jupyter users who have priority
    'ANALY_INFN-T1_GPU': 2 #single queue, no submission parameters, 1 GPU per job
}

kDefultGridSiteType  = ['analysis', 'unified']
kDefaultGridSiteInfo = ['state', 'status', 'maxinputsize', 'maxmemory', 'maxtime']

#########################################################
#############    Configuration Defaults    ##############
#########################################################

kDefaultContainer = '/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:latest'
kDefaultContainer2 = '/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/clcheng/hyperparameter-optimization-on-the-grid:latest'
kDefaultiDDSContainer = 'gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:latest'
kDefaultContainer4 = 'docker://gitlab-registry.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid:base'

kSearchAlgorithms = AlgorithmType.get_members()
kDefaultSearchAlgorithm = 'random'

kAlgoMap = {
    'random': 'tune',
    'bayesian': 'hyperopt'
}

kSchedulers = SchedulerType.get_members()
kDefaultScheduler = 'asynchyperband'

kGenerators = GeneratorMethod.get_members()
kDefaultGenerator = 'nevergrad'
kDefaultMetric = 'loss'
kMetricMode = ['min', 'max']
kDefaultMetricMode = 'min'
kDefaultTrials = 100
kDefaultMaxConcurrent = 3
kDefaultStopping = {'training_iteration': 1}
kDefaultResource = {"cpu": 0, "gpu": 0}
kDefaultLogDir = None
kDefaultVerbosity = 0
kDefaultGridExtraParam  = {}
kDefaultSchedulerParam  = None
kDefaultAlgorithmParam  = None
kDefaultModelParam      = {}

kDefaultOutDS = 'user.${{RUCIO_ACCOUNT}}.hpogrid.{HPO_PROJECT_NAME}.out.$(date +%Y%m%d%H%M%S)'

kDefaultExtraColumns = []


kSearchSpaceFormat = {
    'categorical': {
        'categories' : { 'type': list, 'required': True},
        'grid_search' : { 'type': int, 'required': False, 'default': 0}
    },
    'uniform': {
        'low': { 'type': (float, int), 'required': True},
        'high': { 'type': (float, int), 'required': True}
    },
    'uniformint': {
        'low': { 'type': int, 'required': True},
        'high': { 'type': int, 'required': True}
    },
    'quniform': {
        'low': { 'type': (float, int), 'required': True},
        'high': { 'type': (float, int), 'required': True},
        'q': { 'type': (float, int), 'required': True}
    },
    'loguniform': {
        'low': { 'type': (float, int), 'required': True},
        'high': { 'type': (float, int), 'required': True},
        'base': { 'type': (float, int), 'required': False}
    },
    'qloguniform': {
        'low': { 'type': (float, int), 'required': True},
        'high': { 'type': (float, int), 'required': True},
        'q': { 'type': (float, int), 'required': True},
        'base': { 'type': (float, int), 'required': False}
    },
    'normal': {
        'mu': { 'type': (float, int), 'required': True},
        'sigma': { 'type': (float, int), 'required': True}
    },
    'qnormal': {
        'mu': { 'type': (float, int), 'required': True},
        'sigma': { 'type': (float, int), 'required': True},
        'q': { 'type': (float, int), 'required': True}
    },
    'lognormal': {
        'mu': { 'type': (float, int), 'required': True},
        'sigma': { 'type': (float, int), 'required': True},
        'base': { 'type': (float, int), 'required': False}
    },
    'qlognormal': {
        'mu': { 'type': (float, int), 'required': True},
        'sigma': { 'type': (float, int), 'required': True},
        'base': { 'type': (float, int), 'required': False},
        'q': { 'type': (float, int), 'required': True}
    },
    'fixed': {
        'value': { 'type': None, 'required': True}
    }
}

kConfigFormat = {
    ConfigType.PROJECT:{
        'project_name': {
            'required': True,
            'type': str
        },
        'scripts_path': {
            'abbr': 'p',
            'description': 'Path to the location of training scripts ' +\
                           ' (or the directory containing the training scripts)',
            'required': True,
            'type': str
        },
        'model_config': {
            'abbr': 'm',
            'description': 'Name of the model configuration to use',
            'required': True,
            'type': dict
        },
        'search_space': {
            'abbr': 's',
            'description': 'Name of the search space configuration to use',
            'required': True,
            'type':dict
        },
        'hpo_config': {
            'abbr': 'o',
            'description': 'Name of the hpo configuration to use',
            'required': True,
            'type':dict
        },
        'grid_config': {
            'abbr': 'g',
            'description': 'Name of the grid configuration to use',
            'required': True,
            'type':dict
        }        
    },
    ConfigType.HPO:{
        'algorithm': {
            'abbr': 'a',
            'description': 'Algorithm for hyperparameter optimization',
            'required': False,
            'type': str,
            'choice': kSearchAlgorithms,
            'default': kDefaultSearchAlgorithm
        },
        'scheduler': {
            'abbr': 's',
            'description': 'Trial scheduling method for hyperparameter optimization',
            'required': False,
            'type': str,
            'choice': kSchedulers,
            'default': kDefaultScheduler
        },        
        'metric' : {
            'abbr': 'm',
            'description': 'Evaluation metric to be optimized',
            'required': False,
            'type': str,
            'default': kDefaultMetric
        },
        'extra_metrics': {
            'abbr': 'e',
            'description': 'Additional metrics to be saved during the training (separated by commas)',
            'required': False,
            'type': (list, type(None)),
            'default': None
        },
        'mode' : {
            'abbr': 'o',
            'description': 'Mode of optimization (either "min" or "max")',
            'required': False,
            'type': str,
            'choice': kMetricMode,
            'default': kDefaultMetricMode
        },
        'resource': {
            'abbr': 'r',
            'description': 'A json decodable string defining the resource allocated to each trial, use all cpu and gpu by default',
            'required': False,
            'type': (dict, type(None)),
            'default': deepcopy(kDefaultResource)
        },
        'num_trials': {
            'abbr': 'n',
            'description': 'Number of trials (search points) to run',
            'required': True,
            'type': int,
        },
        'log_dir': {
            'abbr': 'l',
            'description': 'Directory for saving Ray Tune logs, default is /tmp/',
            'required': False,
            'type': (str, type(None)),
            'default': kDefaultLogDir
        },
        'verbose': {
            'abbr': 'v',
            'description': 'Verbosity level of Ray Tune',
            'required': False,
            'type': int,
            'default': kDefaultVerbosity
        },
        'max_concurrent': {
            'abbr': 'c',
            'description': 'Maximum number of trials to be run concurrently',
            'required': False,
            'type': int,
            'default': kDefaultMaxConcurrent
        },
        'stop': {
            'abbr': None,
            'description': 'A json decodable string defining the stopping criteria for the training',
            'required': False,
            'type': dict,
            'default': deepcopy(kDefaultStopping)
        },
        'scheduler_param': {
            'abbr': None,
            'description': 'A json decodable string defining the extra parameters given to the trial scheduler',
            'required': False,
            'type': (dict, type(None)),
            'default': None
        },
        'algorithm_param': {
            'abbr': None,
            'description': 'A json decodable string defining the extra parameters given to the hyperparameter optimization algorithm',
            'required': False,
            'type': (dict, type(None)),
            'default': None
        }
    },
    ConfigType.GRID: {
        'site': {
            'abbr': 's',
            'description': 'Name of the grid site(s) to where the jobs are submitted',
            'required': False,
            'type': (list, str, type(None)),
            'default': None
        },
        'container': {
            'abbr': 'c',
            'description': 'Name of the docker or singularity container in which the jobs are run',
            'required': False,
            'type': str,
            'default': kDefaultContainer
        },
        'inDS': {
            'abbr': 'i',
            'description': 'Name of (rucio) input dataset', 
            'required': False,
            'type': (str, type(None)),
            'default': None
        },
        'outDS': {
            'abbr': 'o',
            'description': 'Name of output dataset',
            'required': False,
            'type': str,
            'default': kDefaultOutDS
        },            
        'retry': {
            'abbr': 'r',
            'description': 'Check to enable retrying faild jobs',
            'required': False,
            'type': (bool, int),
            'default': 0
        },
        'extra': {
            'abbr': 'e',
            'description': 'Extra options passed to panda command',
            'required': False,
            'type': dict,
            'default': deepcopy(kDefaultGridExtraParam)
        }
    },
    ConfigType.MODEL:{
        'script': {
            'abbr': 's',
            'description': 'Name of the training script where the function or class that defines'+\
                           ' the training model will be called to perform the training',
            'required': True,
            'type': str
        },
        'model': {
            'abbr': 'm',
            'description': 'Name of the function or class that defines the training model',
            'required': True,
            'type': str
        },
        'param': {
            'abbr': 'p',
            'description': 'Extra parameters to be passed to the training model',
            'required': False,
            'type': dict,
            'default': deepcopy(kDefaultModelParam)
        }
    },
    ConfigType.SEARCH_SPACE:{
        'search_space': {
            'abbr': 's',
            'description': 'A json decodable string defining the search space',
            'required': True,
            'type': dict,
        }
    }
}

kHPOGridMetadataFormat = ['task_time_s', 'start_timestamp', 'project_name', 
                          'result', 'hyperparameters', 'metric', 'start_datetime', 
                          'best_config', 'best_metric_score', 'end_datetime', 'mode']

kGridSiteMetadataFileName = 'userJobMetadata.json' #jobReport.json

#########################################################
###############    PanDA Tasks Related    ###############
#########################################################

kPanDATaskOutputColumns = ['jeditaskid', 'status', 'taskname', 'computingsite', 'site', 'metastruct']


from hpogrid._version import __version__
from quickstats.decorators import *

import os
import pathlib

module_path = pathlib.Path(__file__).parent.absolute()
resource_path = os.path.join(module_path, "resources")
userdata_path = os.path.join(module_path, "userdata")

from hpogrid.utils.io import  VerbosePrint
stdout = VerbosePrint("INFO")

from hpogrid.core.environment_settings import get_datadir, get_workdir


from .base_space import BaseSpace

class AxSpace(BaseSpace):

    def __init__(self, search_space = None):
        self.library = 'ax'
        super().__init__(search_space)
        
    def reset_space(self):
        self.search_space = []

    def append(self, space_value):
        self.search_space.append(space_value)
    
    def _range(self, name, bounds, value_type = "bounds", log_scale = False):
        space_value = {
            "name": name,
            "type": "range",
            "bounds": bounds,
            "value_type": value_type,
            "log_scale": log_scale
        }
        return space_value


    def categorical(self, label, categories, grid_search = False):
        space_value = {
            "name": label,
            "type": "choice",
            "values": categories,
            "is_ordered": False
        }
        if all(isinstance(value, str) for value in categories):
            space_value['value_type'] = "str"
        elif all(isinstance(value, bool) for value in categories):
            space_value['value_type'] = "bool"
        elif all(isinstance(value, int) for value in categories):
            space_value['value_type'] = "int"
        else:
            space_value['value_type'] = "float"            
        return space_value

    def uniform(self, label, low, high):
        return self._range(label, [low, high], "float")

    def uniformint(self, label, low, high):
        return self._range(label, [low, high], "int")

    def loguniform(self, label, low, high, base = 10):
        if base != 10:
            raise ValueError(f'{self.library} search space only allows base 10 for loguniform sampling')
        return self._range(label, [low, high], "float", log_scale=True)

    def qloguniform(self, label, low, high, q = 1):
        if q != 1:
            raise ValueError(f'{self.library} search space only allows q = 1 for qloguniform sampling')
        return self._range(label, [low, high], "int", log_scale=True)


    def fixed(name, label, value):
        space_value = {
            "name": label,
            "value": value
        }
        return space_value


import os
import sys
from .defaults import kGridSiteWorkDir, RunMode

def get_run_mode():
    environ_val = os.environ.get('HPOGRID_RUN_MODE', 'LOCAL')
    return RunMode.parse(environ_val)

def get_datadir():
    return os.environ.get('HPOGRID_DATA_DIR', os.getcwd())

def get_workdir():
    return os.environ.get('HPOGRID_WORK_DIR', os.getcwd())

def get_grid_workdir():
    if os.path.exists(kGridSiteWorkDir):
        return kGridSiteWorkDir
    else:
        return os.getcwd()

def is_grid_job():
    return (get_run_mode() in [RunMode.GRID, RunMode.IDDS])

def is_idds_job():
    return (get_run_mode() == RunMode.IDDS)

def is_local_job():
    return (get_run_mode() == RunMode.LOCAL)
    
def setup(run_mode:str='local', silent:bool=False):
    from hpogrid import stdout
    if silent:
        print_msg = stdout.error
    else:
        print_msg = stdout.info
    print_msg('INFO: Setting up HPO task')
    
    run_mode = RunMode.parse(run_mode).name
    
    os.environ['HPOGRID_RUN_MODE'] = run_mode
    
    _run_mode = RunMode.parse(run_mode)
    if _run_mode == RunMode.LOCAL:
        os.environ['HPOGRID_DATA_DIR'] = os.getcwd()
        os.environ['HPOGRID_WORK_DIR'] = os.getcwd()
    elif _run_mode == RunMode.GRID:
        os.environ['HPOGRID_DATA_DIR'] = get_grid_workdir()
        os.environ['HPOGRID_WORK_DIR'] = get_grid_workdir()
    elif _run_mode == RunMode.IDDS:
        os.environ['HPOGRID_DATA_DIR'] = get_grid_workdir()
        os.environ['HPOGRID_WORK_DIR'] = get_grid_workdir()
    else:
        raise ValueError(f"unknown run mode: {run_mode.lower()}")
    print_msg(f'INFO: HPO task will run in {run_mode.lower()} environment')
    print_msg('INFO: Work directory is set to "{}"'.format(get_workdir()))
    print_msg('INFO: Data directory is set to "{}"'.format(get_datadir()))
    
def reset():
    os.environ.pop('HPOGRID_RUN_MODE', None)
    os.environ.pop('HPOGRID_DATA_DIR', None)
    os.environ.pop('HPOGRID_WORK_DIR', None)
    
def set_scripts_path(scripts_path:str, undo:bool=False):
    from hpogrid import stdout
    if (scripts_path in sys.path) and undo:
        stdout.info('INFO: Removed {} from $PYTHONPATH'.format(scripts_path))
        sys.path.remove(scripts_path)
        os.environ["PYTHONPATH"].replace(scripts_path+":","")
        
    if (scripts_path not in sys.path) and (not undo):
        stdout.info('INFO: Adding {} to $PYTHONPATH'.format(scripts_path))
        sys.path.append(scripts_path)
        os.environ["PYTHONPATH"] = scripts_path + ":" + os.environ.get("PYTHONPATH", "")

def get_scripts_path(proj_name:str):
    project_path = get_project_path(proj_name)
    scripts_path = os.path.join(project_path, 'scripts')
    return scripts_path
        
def set_scripts_path_from_project(proj_name:str, undo:bool=False):
    project_path = get_project_path(proj_name)
    scripts_path = os.path.join(project_path, 'scripts')
    set_scripts_path(scripts_path, undo=undo)       


import os
import click

from hpogrid.core.defaults import *

from hpogrid.clis.core import ListOption

@click.group(name='tasks')
def manage_tasks(**kwargs):
    """
    Monitor grid jobs
    """
    pass

@manage_tasks.command(name='show')
@click.option('-u', '--username', default=None, 
              help='filter tasks by username')
@click.option('-l', '--limit', type=int, default=1000,
              help='maximum number of tasks to query', show_default=True)
@click.option('-d', '--days', type=int, default=30,
              help='filter tasks within the last N days', show_default=True)
@click.option('-n', '--taskname', default=None,
              help='filter tasks by taskname (accept wildcards)')
@click.option('-r', '--range', 'taskid_range', default=None,
              help='filter tasks by jeditaskid range')
@click.option('-i', '--jeditaskid', default=None,
              help='only show the task with the specified jeditaskid (accept wildcards)')
@click.option('-m', '--metadata', is_flag=True,
              help='print out the metadata of a task')
@click.option('-s', '--status', cls=ListOption, default=[],
              help='filter tasks by task status, separated by commas, '
              'e.g. running,finished,done')
@click.option('--sync', is_flag=True,
              help='force no caching on the PanDA server')
@click.option('-o', '--outname', default=None,
              help='output result with the filename if specified')
@click.option('-c', '--outcol', cls=ListOption, default=kPanDATaskOutputColumns,
              help='data columns to be saved in output, seperated by commas')
def show_task(**kwargs):
    """
    Show an overview of task status
    """
    from hpogrid.interface.panda import PandaTaskManager
    PandaTaskManager(skip_pbook=True).show(**kwargs)
    
@manage_tasks.command(name='kill', short_help="Kill a PanDA task")
@click.argument('ID', type=int)
def kill_task(**kwargs):
    """
    Kill a task
    
    ID: JEDI task ID
    """
    from hpogrid.interface.panda import PandaTaskManager
    kwargs['task_id'] = kwargs.pop('id')
    PandaTaskManager().kill(**kwargs)

@manage_tasks.command(name='retry', short_help="Retry a PanDA task")
@click.argument('ID', type=int)
def retry_task(**kwargs):
    """
    Retry a task
    
    ID: JEDI task ID
    """    
    from hpogrid.interface.panda import PandaTaskManager
    kwargs['task_id'] = kwargs.pop('id')
    PandaTaskManager().retry(**kwargs)
    
@manage_tasks.command(name='kill_and_retry', short_help="Kill and retry a PanDA task")
@click.argument('ID', type=int)
def kill_and_retry_task(**kwargs):
    """
    Kill and retry a task
    
    ID: JEDI task ID
    """      
    from hpogrid.interface.panda import PandaTaskManager
    kwargs['task_id'] = kwargs.pop('id')
    PandaTaskManager().killAndRetry(**kwargs)

    
@click.command(name='run', short_help='Run an HPO task')
@click.argument('INPUT')
@click.option('-p', '--search_points', default=None,
              help='A json file containing the manual search points to run')
@click.option('-m', '--mode', default='local', type=click.Choice(['local', 'grid', 'idds'], case_sensitive=False),
              help='Platform for running hyperparameter optimization', show_default=True)
def run_tasks(**kwargs):
    """
    Run an HPO task
    
    INPUT: Name of project / Path to a project configuration file    
    """
    kwargs['source'] = kwargs.pop('input')    
    from hpogrid.components import HPOTask
    job = HPOTask.load(**kwargs)
    job.run() 
    
@click.command(name='submit', short_help='Submit an HPO task to the grid/idds',
               context_settings=dict(ignore_unknown_options=True, allow_extra_args=True,))
@click.argument('INPUT')
@click.option('-n', '--n_jobs', type=int, default=1,
              help='number of (repeated) jobs to submit', show_default=True)
@click.option('-s', '--site', default=None,
              help='site to submit the job to '
            '(this will override the grid config site setting)')
@click.option('-t', '--time', type=int, default=-1,
              help='same as maxCpuCount in prun which specifies the maximum cpu time'
                   ' for a job (prevent being killed by looping job detection)', show_default=True)
@click.option('-m', '--mode', type=click.Choice(['grid', 'idds'], case_sensitive=False), default='grid',
              help='mode for job submission', show_default=True)
@click.option('-p', '--search_points', default=None,
              help='json file containing a list of '
                   'search points to evaluate (for non-idds jobs only)')
@click.option('-d', '--debug', is_flag=True,
              help='activate debug mode')
@click.option('--scripts_path', default=None,
              help='path to the training script or directory containing the training scripts')
@click.pass_context
def submit_tasks(ctx, **kwargs):
    """
    Submit an HPO task to the grid/idds
    
    INPUT: Name of project / Path to a project configuration file
    """
    kwargs['config_input'] = kwargs.pop('input')
    from hpogrid import GridHandler
    GridHandler().submit(**kwargs, extra=ctx.args)
    
    
@click.command(name='report_grid')
@click.argument('project_name')
@click.option('-l', '--limit', type=int, default=1000,
              help='maximum number of tasks to query', show_default=True)
@click.option('-d', '--days', type=int, default=30,
              help='filter tasks within the last N days', show_default=True)
@click.option('-n', '--taskname', default=None,
              help='filter tasks by taskname (accept wildcards)')
@click.option('-r', '--range', 'taskid_range', default=None,
              help='filter tasks by jeditaskid range')
@click.option('-e', '--extra', 'extra_columns', default='',
              help='extra data columns (separated by commas) to be displayed and saved. '
                   'Available options: site, task_time_s, time_s, taskid, start_timestamp, '
                   'start_datetime, end_datetime', show_default=True)
@click.option('--summary', is_flag=True, 
              help='save a complete hpo result summary as a json file')
@click.option('--to_json', is_flag=True,
              help='output result to a json file')
@click.option('--to_html', is_flag=True,
              help='output result to an html file')
@click.option('--to_csv', is_flag=True,
              help='output result to a csv file')
@click.option('--to_pcp', is_flag=True,
              help='output result as a parallel coordinate plot')
@click.option('--to_mlflow', is_flag=True,
              help='output result as an mlflow tracker directory')
@click.option('-o', '--outname', default='hpo_result',
              help='output file name (excluding extension)', show_default=True)
@click.option('--min', 'metric_min', type=float, default=None,
              help='filter results by minimum value of metric')
@click.option('--max', 'metric_max', type=float, default=None,
              help='filter results by maximum value of metric')        
def report_grid(**kwargs):
    """
    Display a report of an HPO task that ran on grid
    """
    from hpogrid.interface.panda import PandaTaskReport
    kwargs['extra_columns'] = [i.strip() for i in kwargs['extra_columns'].split(',') if i.strip()]
    PandaTaskReport().display(**kwargs)
    
@click.command(name='report')
@click.option('-e', '--extra', 'extra_columns', default='',
              help='extra columns to included (separated by comma)')
@click.option('--min', 'metric_min', type=float, default=None,
              help='filter results by minimum value of metric')
@click.option('--max', 'metric_max', type=float, default=None,
              help='filter results by maximum value of metric')
@click.argument('filename')     
def report(**kwargs):
    """
    Display a report of an HPO task that ran locally
    """
    from hpogrid.components import HPOReport
    filename = kwargs.pop('filename')
    if not os.path.exists(filename):
        raise FileNotFoundError(f"input file {filename} does not exist")
    report  = HPOReport.from_json(filename)
    kwargs['extra_columns'] = [i.strip() for i in kwargs['extra_columns'].split(',') if i.strip()]
    report.show()


import os
import sys
import glob
import json
import time
import tempfile
import importlib
from copy import deepcopy
from typing import Optional, Dict, List, Callable, Union, Tuple
from datetime import datetime

import numpy as np
import pandas as pd

from hpogrid.core.defaults import *
from hpogrid.core import config_manager
from hpogrid.utils import helper
import hpogrid.core.environment_settings as es

from .abstract_object import AbstractObject

class HPOTask(AbstractObject):
    
    @property
    def algorithm(self) -> str:
        return self._algorithm
    
    @algorithm.setter
    def algorithm(self, name):
        if name and (name not in kSearchAlgorithms):
            raise ValueError('unrecognized search algorithm: {}'.format(name))
        self._algorithm = kAlgoMap.get(name, name)
    
    @property
    def scheduler(self) -> str:
        return self._scheduler
    
    @scheduler.setter
    def scheduler(self, name):
        if name and (name not in kSchedulers):
            raise ValueError('unrecognized trial scheduler: {}'.format(name))
        self._scheduler = name

    @property
    def mode(self) -> str:
        return self._mode
    
    @mode.setter
    def mode(self, value):
        if value not in kMetricMode:
            raise ValueError('metric mode must be either "min" or "max"')
        self._mode = value
    
    @property
    def resource(self):
        return self._resource
    
    @property
    def resource_per_trial(self):
        return self._resource_per_trial
    
    @resource_per_trial.setter
    def resource_per_trial(self, value):
        if not self.resource:
            self._resource = self.get_resource_info()
        resource_per_trial = {}
        num_trials = self.num_trials if self.num_trials != -1 else max(self.resource['gpu'], 1)
        max_concurrent = min(num_trials, self.max_concurrent)
        for device in ['cpu', 'gpu']:
            n_device = self.resource[device]
            if n_device >= max_concurrent:
                resource_per_trial[device] = int(n_device/max_concurrent)  
            elif n_device > 0:
                resource_per_trial[device] = n_device / max_concurrent
            else:
                resource_per_trial[device] = 0
        # for now only use 1 GPU per trial
        if self.num_trials == -1:
            resource_per_trial['gpu'] = 1
            
        if isinstance(value, dict):
            for device in ['cpu', 'gpu']:
                if device in value and value[device]:
                    resource_per_trial[device] = min(value[device], resource_per_trial[device])
        self._resource_per_trial = resource_per_trial
    
    @property
    def algorithm_param(self):
        return self._algorithm_param
    
    @property
    def scheduler_param(self):
        return self._scheduler_param
    
    @algorithm_param.setter
    def algorithm_param(self, val):
        self._algorithm_param = val if val is not None else {}
            
    @scheduler_param.setter
    def scheduler_param(self, val):
        self._scheduler_param = val if val is not None else {}
        
    @property
    def model(self):
        return self._model
    
    @property
    def search_space(self):
        return self._search_space   
        
    @property
    def stop(self):
        return self._stop
    
    @stop.setter
    def stop(self, val):
        self._stop = val if val is not None else {}    
        
    @property
    def df(self) -> 'pandas.DataFrame':
        return self._df
    
    @property
    def start_datetime(self) -> str:
        return self._start_datetime
    
    @property
    def end_datetime(self) -> str:
        return self._end_datetime
    
    @property
    def start_timestamp(self) -> float:
        return self._start_timestamp
    
    @property
    def total_time(self) -> float:
        return self._total_time
    
    @property
    def best_config(self) -> Dict:
        return self._best_config
    
    @property
    def best_metric_score(self) -> Dict:
        return self._best_metric_score

    @property
    def hyperparameters(self):
        return self._hyperparameters
    
    @property
    def run_mode(self):
        return self._run_mode
    
    @property
    def log_dir(self) -> str:
        return self._log_dir

    @log_dir.setter
    def log_dir(self, val):
        """
        self._log_dir = val
        if hpogrid.is_local_job() and (not os.path.isabs(val)):
            project_path = helper.get_project_path(self.project_name)
            if os.path.exists(project_path):
                self._log_dir = os.path.join(project_path, val)
        self._log_dir = helper.pretty_path(self._log_dir)
        """
        self._log_dir = val    
    
    def __init__(self, model:Optional[Callable]=None, search_space:Optional[Dict]=None,
                 verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)
        self._resource = None        
        self.reset()
        self.set_model(model)
        self.set_search_space(search_space)
        self.set_run_mode("local")
        
    def setup(self, project_name:Optional[str]=None, algorithm:str=kDefaultSearchAlgorithm,
              metric:str=kDefaultMetric, mode:str=kDefaultMetricMode, num_trials:int=kDefaultTrials,
              max_concurrent:int=kDefaultMaxConcurrent, stop:Optional[Dict]=None,
              scheduler:str=kDefaultScheduler, resource_per_trial:Optional[Dict]=None,
              verbose:int=kDefaultVerbosity, log_dir:Optional[str]=kDefaultLogDir,
              extra_metrics:Optional[Dict]=None, algorithm_param:Optional[Dict]=None,
              scheduler_param:Optional[Dict]=None, model_script:Optional[str]=None,
              model_name:Optional[str]=None, model_param:Optional[Dict]=None,
              scripts_path:Optional[str]=None):
        
        # set default values for dict type arguments
        if stop is None:
            stop = deepcopy(kDefaultStopping)
        if algorithm_param is None:
            algorithm_param = deepcopy(kDefaultAlgorithmParam)
        if scheduler_param is None:
            scheduler_param = deepcopy(kDefaultSchedulerParam)
        if resource_per_trial is None:
            resource_per_trial = deepcopy(kDefaultResource)
        if model_param is None:
            model_param = {}
        if scripts_path is None:
            scripts_path = os.getcwd()
            
        self.project_name       = project_name
        self.algorithm          = algorithm
        self.metric             = metric
        self.mode               = mode
        self.num_trials         = num_trials
        self.max_concurrent     = max_concurrent
        self.stop               = stop
        self.scheduler          = scheduler
        self.resource_per_trial = resource_per_trial
        self.verbose            = verbose
        self.log_dir            = log_dir
        self.extra_metrics      = extra_metrics
        self.scheduler_param    = scheduler_param
        self.algorithm_param    = algorithm_param
        self.model_script       = model_script
        self.model_name         = model_name
        self.model_param        = model_param
        self.scripts_path       = scripts_path
        
    def export_setup(self):
        attributes = ["project_name", "algorithm", "metric", "mode",
                      "num_trials", "max_concurrent", "stop", "scheduler",
                      "resource_per_trial", "verbose", "log_dir", "extra_metrics",
                      "scheduler_param", "algorithm_param", "model_script",
                      "model_name", "model_param", "scripts_path"]
        result = {}
        for attribute in attributes:
            result[attribute] = self.__getattribute__(attribute)
        return result

    def reset(self):
        self.set_model()
        # hpo configuration
        self.setup()
        # search space
        self.set_search_space()
        # parameters for logging
        self.outdir = "./"
        self._df = None
        self._start_datetime = None
        self._end_datetime = None
        self._total_time = None
        self._best_config = None
        self._best_metric_score = None
        self._start_timestamp = None
        self.idds_job = False
        self.search_points = None
        self._df_restore = None
        
    def set_model(self, model:Optional[Callable]=None):
        self._model = model
        
    def set_search_space(self, search_space:Optional[Dict]=None):
        if search_space is None:
            self._search_space = None
            self._hyperparameters = []
        elif isinstance(search_space, dict):
            self._search_space = deepcopy(search_space)
            self._hyperparameters = list(self.search_space.keys())
        else:
            raise ValueError("search space must be of type \"dict\"")
            
    def set_search_points(self, search_points:Union[str, List]):
        if isinstance(search_points, str):
            with open(search_points, 'r') as f:
                search_points = json.load(f)
        self.search_points = search_points
        from hpogrid.search_space import SkOptSpace
        skopt_search_points = SkOptSpace.transform(search_points, reference=self.search_space)
        self.algorithm = 'skopt'
        self.scheduler = None
        self.scheduler_param = {}
        self.algorithm_param = {'points_to_evaluate': skopt_search_points}
        self.num_trials = len(skopt_search_points)
        self.stdout.info("INFO: Using explicit search points for execution. "
                         "Search algorithm is set to \"skopt\" and trial "
                         "scheduler is set to None. Parameters for scheduler "
                         "and search algorithm is reset. Number of trials is changed "
                         "according to the number of search points provided")            
            
    @staticmethod
    def get_scheduler(name:str, metric:str, mode:str, search_space=None, **args):
        if (name == None) or (name == 'None'):
            return None
        _name = SchedulerType.parse(name)
        if _name == SchedulerType.ASYNCHYPERBAND:
            from hpogrid.schedulers import AsyncHyperBandScheduler
            return AsyncHyperBandScheduler().create(metric, mode, **args)
        elif _name == SchedulerType.BOHBHYPERBAND:
            from hpogrid.schedulers import BOHBHyperBandScheduler
            return BOHBHyperBandScheduler().create(metric, mode, **args)
        elif _name == SchedulerType.PBT:
            from hpogrid.schedulers import PBTScheduler
            if search_space is None:
                raise ValueError('missing search space definition for pbt scheduler')
            return PBTScheduler().create(metric, mode, search_space, **args)
        else:
            raise ValueError(f'Unrecognized scheduler method: {name}')

    @staticmethod
    def get_search_space(base_search_space:Dict, algorithm:str):
        if base_search_space is None:
            raise ValueError('search space can not be empty')
        _algorithm = SearchSpaceType.parse(algorithm)
        if _algorithm == SearchSpaceType.AX:
            from hpogrid.search_space import AxSpace
            return AxSpace(base_search_space).get_search_space()
        elif _algorithm == SearchSpaceType.BOHB:
            from hpogrid.search_space import BOHBSpace
            return BOHBSpace(base_search_space).get_search_space()
        elif _algorithm == SearchSpaceType.HYPEROPT:
            from hpogrid.search_space import HyperOptSpace
            return HyperOptSpace(base_search_space).get_search_space()
        elif _algorithm == SearchSpaceType.SKOPT:
            from hpogrid.search_space import SkOptSpace
            return SkOptSpace(base_search_space).get_search_space()
        elif _algorithm == SearchSpaceType.TUNE:
            from hpogrid.search_space import TuneSpace
            return TuneSpace(base_search_space).get_search_space()
        elif _algorithm == SearchSpaceType.NEVERGRAD:
            from hpogrid.search_space import NeverGradSpace
            return NeverGradSpace(base_search_space).get_search_space()
        elif _algorithm == SearchSpaceType.OPTUNA:
            from hpogrid.search_space import OptunaSpace
            return OptunaSpace(base_search_space).get_search_space()         
        else:
            raise ValueError(f'Unrecognized search algorithm: {algorithm}')

    @staticmethod
    def get_algorithm(name:str, metric:str, mode:str, search_space:Dict,
                      max_concurrent:int=None, restore_points:Optional[Tuple]=None,
                      **kwargs):
        _name = AlgorithmType.parse(name)
        if _name == AlgorithmType.AX:
            from hpogrid.algorithms import AxAlgorithm
            algo_cls = AxAlgorithm
        elif _name == AlgorithmType.BOHB:
            from hpogrid.algorithms import BOHBAlgorithm
            algo_cls = BOHBAlgorithm
        elif _name == AlgorithmType.HYPEROPT:
            from hpogrid.algorithms import HyperOptAlgorithm
            algo_cls = HyperOptAlgorithm
        elif _name == AlgorithmType.SKOPT:
            from hpogrid.algorithms import SkOptAlgorithm
            algo_cls = SkOptAlgorithm
        elif _name == AlgorithmType.NEVERGRAD:
            from hpogrid.algorithms import NeverGradAlgorithm
            algo_cls = NeverGradAlgorithm
        elif _name == AlgorithmType.OPTUNA:
            from hpogrid.algorithms import OptunaAlgorithm
            algo_cls = OptunaAlgorithm
        elif _name == AlgorithmType.TUNE:
            return None
        else:
            raise ValueError(f'Unrecognized search algorithm: {name}')
        algo_wrapper = algo_cls(search_space=search_space, metric=metric, mode=mode)
        algorithm = algo_wrapper.create(max_concurrent=max_concurrent, 
                                        restore_points=restore_points, **kwargs)
        return algorithm

    def load_model(self, script_name:str, model_name:str, scripts_path:Optional[str]=None):
        scripts_path = os.getcwd() if scripts_path is None else scripts_path
        model = None
        script_name_noext = os.path.splitext(script_name)[0]
        try: 
            es.set_scripts_path(scripts_path)
            module = importlib.import_module(script_name_noext)
            model = getattr(module, model_name)
        except: 
            raise ImportError(f'Unable to import function/class {model_name} '
                              'from training script: {script_name_noext}.py')
        finally:
            es.set_scripts_path(scripts_path, undo=True)
        self.stdout.info('INFO: Loaded module {}'.format(model.__name__))
        return model

    def _process_tune_df(self, df:"pandas.DataFrame"):
        rename_cols = { 'config/{}'.format(hp): hp for hp in self.hyperparameters}
        rename_cols['time_total_s'] = 'time_s'
        df = df.rename(columns=rename_cols)
        cols_to_save = ['time_s'] + self.hyperparameters
        cols_to_save.append(self.metric)
        if self.extra_metrics:
            cols_to_save += self.extra_metrics
        cols_to_save = list(set(cols_to_save))
        df = df.filter(cols_to_save, axis=1)
        return df
    
    def get_processed_df(self, df:Optional["pandas.DataFrame"]=None):
        if df is not None:
            df = self._process_tune_df(df)
            if self._df_restore is not None:
                df = pd.concat([df, self._df_restore], ignore_index=True)
        elif self._df_restore is not None:
            df = self._df_restore
        else:
            raise RuntimeError("no metadata to restore")
        return df
        
    def create_metadata(self, df:Optional["pandas.DataFrame"]=None) -> Dict:
        # if None: use cached result only

        summary = {
            'project_name' : self.project_name,
            'start_datetime': self.start_datetime,
            'end_datetime': self.end_datetime,
            'start_timestamp': self.start_timestamp,
            'task_time_s' : self.total_time,
            'hyperparameters': self.hyperparameters,
            'metric': self.metric,
            'mode' : self.mode,
            'best_config': self.best_config,
            'best_metric_score': self.best_metric_score
        }
         
        df = self.get_processed_df(df)
        df = df.transpose()
        
        summary['result'] = df.to_dict()
        
        return summary

    def get_resource_info(self) -> Dict:
        resource = {}
        n_gpu = helper.get_n_gpu()
        n_cpu = helper.get_n_cpu()
        resource['gpu'] = n_gpu
        resource['cpu'] = n_cpu
        return resource
    
    def print_resource_info(self):
        self.stdout.info('INFO: Number of GPUs detected: {}'.format(self.resource['gpu']))
        self.stdout.info('INFO: Number of CPUs detected: {}'.format(self.resource['cpu']))
        self.stdout.info('INFO: Each trial will use {} GPU(s) resource'.format(self.resource_per_trial['gpu']))
        self.stdout.info('INFO: Each trial will use {} CPU(s) resource'.format(self.resource_per_trial['cpu']))
    
    def print_setup(self):
        from hpogrid.utils.stylus import create_formatted_dict
        text = create_formatted_dict(self.export_setup(), indexed=False, columns=["Attribute","Value"])
        self.stdout.info(text)
        
    def get_all_metrics(self):
        all_metrics = [self.metric]
        if self.extra_metrics:
            for extra_metric in self.extra_metrics:
                if extra_metric not in all_metrics:
                    all_metrics.append(extra_metric)
        return all_metrics
    
    def run(self, outdir:Optional[str]=None, cache:bool=True) -> None:
            
        self._start_timestamp = start = time.time()
        self._start_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        
        es.setup(self.run_mode)
        
        # get model parameters
        if self.algorithm == 'tune':
            tune_config_space = self.get_search_space(self.search_space, algorithm='tune')
        else:
            tune_config_space = {}

        tune_config_space.update(self.model_param)
                
        # load model from path
        if (not self.model):
            model = self.load_model(self.model_script, self.model_name, self.scripts_path)
        else:
            model = self.model
            
        self.print_resource_info()
        
        if outdir is None:
            outdir = self.outdir
        
        metadata_path = os.path.join(outdir, kGridSiteMetadataFileName)
        cache_dir     = os.path.join(outdir, "cache")
        
        if cache and os.path.exists(cache_dir):
            self.stdout.info(f"INFO: Attempting to restore search results from \"{cache_dir}\"")
            restore_points = self._get_restore_points(cache_dir, load_df=True)
        else:
            restore_points = None

        if restore_points:
            n_cached = len(restore_points[0])
            n_restored = len([v for v in restore_points[1] if not helper.is_nan_or_inf(v)])
            self.stdout.info(f"INFO: Found {n_cached} search results from cache")
            self.stdout.info(f"INFO: Will restore {n_restored} valid search results")
        else:
            n_cached = 0
            n_restored = 0

        # get trial scheduler
        scheduler = self.get_scheduler(
            self.scheduler,
            self.metric,
            self.mode,
            self.search_space,
            **self.scheduler_param)
 
        # get callbacks
        callbacks = self.get_callbacks(cache_dir=cache_dir)
        
        # determine the number of search points to run accounting for restored points
        if self.num_trials == -1:
            num_samples = max(self.resource.get('gpu'), 1)
        else:
            num_samples = self.num_trials
            
        if (self.num_trials != -1) and (n_restored > 0):
            num_samples_remain = num_samples - n_restored
        else:
            num_samples_remain = num_samples
                
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        
        # save something to prevent looping job
        with open(metadata_path,'w') as output:
            json.dump({}, output)

        temp_dir = tempfile.TemporaryDirectory()
        
        if not self.log_dir:
            local_dir = os.path.join(temp_dir.name, 'log')
        else:
            local_dir = self.log_dir

        if num_samples_remain > 0:
            
            # get search algorithm
            # move to here so that if jobs are already done we don't need to restore
            algorithm = self.get_algorithm(
                self.algorithm,
                self.metric,
                self.mode,
                self.search_space,
                self.max_concurrent,
                restore_points=restore_points,
                **self.algorithm_param)
            
            import ray
            from ray import tune

            # HPO starts
            try:
                # extract input dataset if it is a tar file
                # (for grid jobs only)
                extracted_files = []
                if es.is_grid_job():
                    datadir = es.get_datadir()
                    extracted_files = helper.extract_tarball(datadir, datadir)
                ray.init(include_dashboard=False)

                if AlgorithmType.parse(self.algorithm) == AlgorithmType.NEVERGRAD:
                    try:
                        algo_str = f"{self.algorithm} ({algorithm.searcher._nevergrad_opt.name})"
                    except:
                        algo_str = f"{self.algorithm}"
                else:
                    algo_str = f"{self.algorithm}"
                
                if num_samples_remain != num_samples:
                    sample_str = f"{num_samples} ({num_samples - num_samples_remain} restored)"
                else:
                    sample_str = f"{num_samples}"

                self.stdout.info(f"INFO: Ray Tune log files will be saved to {local_dir}")
                self.stdout.info("--------------------------------------------------------------------------------")
                self.stdout.info(f"INFO: Begin hyperparameter optimization with the following settings")
                self.stdout.info(f"\tProject Name            : {self.project_name}")
                self.stdout.info(f"\tSearch Algorithm        : {algo_str}")
                self.stdout.info(f"\tTrial Scheduler         : {self.scheduler}")
                self.stdout.info(f"\tNumber of Search Points : {sample_str}")
                self.stdout.info("--------------------------------------------------------------------------------")
                cwd = os.getcwd()
                analysis = tune.run(
                    model,
                    name=self.project_name,
                    scheduler=scheduler,
                    search_alg=algorithm,
                    config=tune_config_space,
                    num_samples=num_samples_remain,
                    resources_per_trial=self.resource_per_trial,
                    verbose=self.verbose,
                    local_dir=local_dir,
                    callbacks=callbacks,
                    raise_on_failed_trial=False,
                    stop=self.stop)
                os.chdir(cwd)
            finally:
                ray.shutdown()
                # remove the extracted tar files
                helper.remove_files(extracted_files)
            
            if num_samples_remain == num_samples:
                best_trial = analysis.get_best_trial(metric=self.metric, mode=self.mode)
                if best_trial:
                    self._best_config = best_trial.config
                    self._best_metric_score = best_trial.metric_analysis[self.metric].get("last", None)
                else:
                    self._best_config = None
                    self._best_metric_score = None
                    
            df = analysis.dataframe(metric=self.metric, mode=self.mode)
        else:
            self.stdout.info(f"INFO: Number of restored points({n_restored}) "
                             f">= number of search points({num_samples}). "
                             "Skip execution.")
            df = None

        if not self.best_config:
            self._best_config, self._best_metric_score = self.get_best_trial_from_tune_df(df)
            
        end = time.time()
        self._total_time = float(end - start)
        self._end_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self.stdout.info(f"Best config: {self.best_config}")
        self.stdout.info(f"Best metric score: {self.best_metric_score}")
        self.stdout.info(f"INFO: Time taken in seconds: {self.total_time}")
        
        metadata = self.create_metadata(df)
        # save metadata
        with open(metadata_path, 'w') as output:            
            json.dump(metadata, output, cls=helper.NpEncoder, indent=2)
            self.stdout.info(f'INFO: Job metadata is saved as {metadata_path}')
        
        # save idds output
        if es.is_idds_job():
            from hpogrid.interface.idds.main import save_idds_output_from_metadata
            save_idds_output_from_metadata(metadata)
            
        # reset environment
        es.reset()
        temp_dir.cleanup()
    
    def get_best_trial_from_tune_df(self, df:Optional["pandas.DataFrame"]=None, metric:Optional[str]=None,
                                    mode:Optional[str]=None, filter_nan_and_inf:bool=True):
        if metric is None:
            metric = self.metric
        if mode is None:
            mode = self.mode
        # validation
        if mode not in ["min", "max"]:
            raise ValueError("optimization mode must be either \"min\" or \"max\"")
            
        best_config = None
        best_metric_score = None
        
        hyperparameters = self.hyperparameters
        
        results = self.get_processed_df(df).transpose().to_dict()
        
        for i, result in results.items():
            
            if metric not in result:
                continue
                
            metric_score = result[metric]
            if filter_nan_and_inf and helper.is_nan_or_inf(metric_score):
                continue
                
            if not set(hyperparameters).issubset(set(result.keys())):
                continue
            config =  {hp:result[hp] for hp in hyperparameters}
            
            if best_metric_score is None:
                best_metric_score = metric_score
                best_config = config
                continue

            if (mode == "max") and (best_metric_score < metric_score):
                best_metric_score = metric_score
                best_config = config
            elif (mode == "min") and (best_metric_score > metric_score):
                best_metric_score = metric_score
                best_config = config

        if not best_config:
            self.stdout.warning("WARNING: Could not find best trial. "
                                "This can be due to invalid \"metric\" parameter "
                                "or all trials ended up in invalid metric values (inf or nan).")
        return best_config, best_metric_score
        
    def _get_restore_points(self, cache_dir:Optional[str]=None, load_df:bool=False):
        if not cache_dir:
            return None
        filenames = glob.glob(os.path.join(cache_dir, "*.json"))
        if not filenames:
            return None

        data = [json.load(open(filename)) for filename in filenames]
        try:
            for _data in data:
                config = _data.pop("config")
                for key, value in config.items():
                    _data[f"config/{key}"] = value
            df = pd.DataFrame(data)
            df = self._process_tune_df(df)
        except:
            raise RuntimeError(f"invalid search results from {cache_dir}")
            
        if load_df:
            self._df_restore = df
            
        hyperparams_values = df[self.hyperparameters].to_dict("records")
        metric_values = df[self.metric].values
        return (hyperparams_values, metric_values)
    
    def get_callbacks(self, cache_dir:Optional[str]=None):
        #from ray.tune.logger import NoopLogger, MLFLowLogger
        #return [NoopLogger, MLFLowLogger]
        from hpogrid.interface.tune import TuneCallback
        tune_callback = TuneCallback()
        hyperparameters = self.hyperparameters
        all_metrics = self.get_all_metrics()
        tune_callback.custom_setup(hyperparameters=hyperparameters,
                                   metrics=all_metrics,
                                   cache_dir=cache_dir)
        return [tune_callback]
    
    def set_run_mode(self, run_mode:str='local'):
        self._run_mode = run_mode

    def parse_project_config(self, source:[Dict, str]):
        
        is_saved_project = config_manager.is_project(source)
        
        config = config_manager.load_configuration(source)
        from hpogrid.configuration import ProjectConfiguration
        config = ProjectConfiguration._validate(config)
        
        # if input is a project name, set scripts path to 
        # where the scripts are stored in the project directory
        # if input is a configuration file, set scripts path to
        # the scripts path stated in configuration file
        if is_saved_project:
            project_name = source
            scripts_path = config_manager.get_project_script_dir(project_name)
        else:
            scripts_path = helper.pretty_dirname(config['scripts_path'])
            
        # Load search space
        self.set_search_space(config['search_space'])
        
        setup_kwargs = {
            "project_name"       : config['project_name'],
            "algorithm"          : config['hpo_config']['algorithm'],
            "metric"             : config['hpo_config']['metric'],
            "mode"               : config['hpo_config']['mode'],
            "scheduler"          : config['hpo_config']['scheduler'],
            "scheduler_param"    : config['hpo_config']['scheduler_param'],
            "algorithm_param"    : config['hpo_config']['algorithm_param'],
            "num_trials"         : config['hpo_config']['num_trials'],
            "max_concurrent"     : config['hpo_config']['max_concurrent'],
            "verbose"            : config['hpo_config']['verbose'],
            "log_dir"            : config['hpo_config'].get('log_dir', None),
            "stop"               : config['hpo_config'].get('log_dir', None),
            "resource_per_trial" : config['hpo_config'].get('resource', None),
            "extra_metrics"      : config['hpo_config'].get('extra_metrics', None),
            "scripts_path"       : scripts_path,
            "model_script"       : config['model_config']['script'],
            "model_name"         : config['model_config']['model'],
            "model_param"        : config['model_config']['param']
        }

        # Load hpo configuration
        self.setup(**setup_kwargs)

    @classmethod
    def load(cls, source:[Dict, str], search_points=None, mode='local'):
        """
        Arguments:
            source: str or dict
                dict: a dict representing the project configuration
                str: path to the json file storing the project configuration, or
                     name of the project saved under hpogrid
        """

        task = cls()
        task.set_run_mode(mode)
        es.setup(mode, silent=True)
        
        task.parse_project_config(source)

        # configuration for idds
        if (not search_points) and es.is_idds_job():
            from hpogrid.interface.idds.main import get_search_points
            search_points = get_search_points()

        if search_points:
            task.set_search_points(search_points)
        
        return task


from hpogrid.search_space.base_space import BaseSpace

class NeverGradSpace(BaseSpace):

    def __init__(self, search_space=None):
        self.library = 'nevergrad'
        super().__init__(search_space)    

    def reset_space(self):
        self.search_space = {}

    def get_search_space(self):
        import nevergrad as ng
        return ng.p.Instrumentation(**self.search_space)

    def append(self, space_value):
        self.search_space[space_value[0]] = space_value[1]

    def categorical(self, label, categories, grid_search = False):
        if grid_search != False:
            raise ValueError(f'{self.library} does not allow grid search')
        import nevergrad as ng
        return (label, ng.p.Choice(categories))

    def uniformint(self, label, low, high):
        import nevergrad as ng
        return (label, ng.p.Scalar(lower=low, upper=high).set_integer_casting())

    def uniform(self, label, low, high):
        import nevergrad as ng
        return (label, ng.p.Scalar(lower=low, upper=high))

    def loguniform(self, label, low, high, base=10):
        import nevergrad as ng
        return (label, ng.p.Log(lower=low, upper=high, exponent=base))

    def loguniformint(self, label, low, high, base=10):
        import nevergrad as ng
        return (label, ng.p.Log(lower=low, upper=high, exponent=base).set_integer_casting())

    def fixed(self, label, value):
        return self.categorical(label=label, categories=[value])



from hpogrid.configuration.configuration_base import ConfigurationBase
from hpogrid.configuration.hpo_configuration import HPOConfiguration
from hpogrid.configuration.grid_configuration import GridConfiguration
from hpogrid.configuration.search_space_configuration import SearchSpaceConfiguration
from hpogrid.configuration.model_configuration import ModelConfiguration
from hpogrid.configuration.project_configuration import ProjectConfiguration


import copy
from typing import List, Dict

from .base_generator import Generator
from hpogrid.search_space.bohb_space import BOHBSpace

class BOHBGenerator(Generator):
    
    kDefaultBudget = 100

    def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
        
        search_space = BOHBSpace(search_space).get_search_space()
        
        from hpbandster.optimizers.config_generators.bohb import BOHB
        searcher = BOHB(search_space)
        
        return searcher

    def _ask(self, searcher, n_points:int = None):
        points = []
        for _ in range(n_points):
            point, info = searcher.get_config(None)
            points.append(copy.deepcopy(point))
        return points

    def _tell(self, searcher, point:Dict, value:float):
        value = self._to_metric_values(value)
        result = BOHBResultWrapper(point, self.signature * value)
        searcher.new_result(result)

class BOHBResultWrapper():
    
    def __init__(self, config:Dict, loss:float, budget:int=BOHBGenerator.kDefaultBudget):
        self.result = {"loss": loss}
        self.kwargs = {"budget": budget, "config": config.copy()}
        self.exception = None


from typing import Dict, Tuple

from hpogrid.search_space import AxSpace
from .base_algorithm import BaseAlgorithm

class AxAlgorithm(BaseAlgorithm):
    
    def __init__(self, metric:str, mode:str, search_space:Dict):
        super().__init__(metric=metric, mode=mode, search_space=search_space)

    def _get_algorithm(self, **args):
        
        search_space = AxSpace(self.search_space).get_search_space()
        
        kwargs = {**args}
        if 'enforce_sequential_optimization' not in kwargs:
            kwargs['enforce_sequential_optimization'] = False
        if 'verbose_logging' not in kwargs:
            kwargs['verbose_logging'] = False

        from ray.tune.suggest.ax import AxSearch
        algorithm = AxSearch(search_space, metric=self.metric, mode=self.mode, **kwargs)
        
        return algorithm
    
    def _restore_search_points(self, algorithm, restore_points:Tuple):
        from hpogrid.generators import AxGenerator
        generator = AxGenerator(search_space=self.search_space,
                                metric=self.metric, mode=self.mode)
        generator.searcher = algorithm._ax
        
        generator.feed(restore_points[0], restore_points[1])


from copy import deepcopy

import click
from click import Option, Argument

from hpogrid.core.defaults import *
from .core import ListOption, DictOption

config_params = {
    ConfigType.HPO: [
        Argument(('NAME',)),
        Option(('-a', '--algorithm',), default=kDefaultSearchAlgorithm,
               type=click.Choice(kSearchAlgorithms, case_sensitive=False),
               help=kConfigFormat[ConfigType.HPO]['algorithm']['description'], show_default=True),
        Option(('-s', '--scheduler',), default=kDefaultScheduler,
               type=click.Choice(kSchedulers, case_sensitive=False),
               help=kConfigFormat[ConfigType.HPO]['scheduler']['description'], show_default=True),
        Option(('-m', '--metric',), default=kDefaultMetric,
               help=kConfigFormat[ConfigType.HPO]['metric']['description'], show_default=True),
        ListOption(('-e', '--extra_metrics',), default=kDefaultMetric,
               help=kConfigFormat[ConfigType.HPO]['extra_metrics']['description'], show_default=True),
        Option(('-o', '--mode',), default=kDefaultMetricMode,
               type=click.Choice(kMetricMode, case_sensitive=False),
               help=kConfigFormat[ConfigType.HPO]['mode']['description'], show_default=True),
        DictOption(('-r', '--resource',), default=None, 
               help=kConfigFormat[ConfigType.HPO]['resource']['description']),
        Option(('-n', '--num_trials',), type=int, required=True,
               help=kConfigFormat[ConfigType.HPO]['num_trials']['description']),
        Option(('-l', '--log_dir',), default=None,
               help=kConfigFormat[ConfigType.HPO]['log_dir']['description']),
        Option(('-v', '--verbose',), default=0, type=int,
               help=kConfigFormat[ConfigType.HPO]['verbose']['description'], show_default=True),
        Option(('-c', '--max_concurrent',), type=int, default=kDefaultMaxConcurrent,
               help=kConfigFormat[ConfigType.HPO]['max_concurrent']['description'], show_default=True),
        DictOption(('--stop',), default=kDefaultStopping,
               help=kConfigFormat[ConfigType.HPO]['stop']['description'], show_default=True),
        DictOption(('--scheduler_param',), default=None,
               help=kConfigFormat[ConfigType.HPO]['scheduler_param']['description']),
        DictOption(('--algorithm_param',), default=None,
               help=kConfigFormat[ConfigType.HPO]['algorithm_param']['description'])
    ],
    ConfigType.GRID:[
        Argument(('NAME',)),
        ListOption(('-s', '--site',), default=None,
               help=kConfigFormat[ConfigType.GRID]['site']['description']),
        Option(('-c', '--container',), default=kDefaultContainer,
               help=kConfigFormat[ConfigType.GRID]['container']['description'], show_default=True),
        Option(('-i', '--inDS', 'inDS',), default=None,
               help=kConfigFormat[ConfigType.GRID]['inDS']['description']),
        Option(('-o', '--outDS', 'outDS',), default=kDefaultOutDS,
               help=kConfigFormat[ConfigType.GRID]['outDS']['description'], show_default=True),
        Option(('-r', '--retry',), is_flag=True,
               help=kConfigFormat[ConfigType.GRID]['retry']['description']),
        DictOption(('-e', '--extra',), default={},
               help=kConfigFormat[ConfigType.GRID]['extra']['description'], show_default=True)
    ],
    ConfigType.MODEL:[
        Argument(('NAME',)),
        Option(('-s', '--script',), required=True,
               help=kConfigFormat[ConfigType.MODEL]['script']['description']),
        Option(('-m', '--model',), required=True,
               help=kConfigFormat[ConfigType.MODEL]['model']['description']),
        DictOption(('-p', '--param',), default={},
               help=kConfigFormat[ConfigType.MODEL]['param']['description'], show_default=True)
    ],
    ConfigType.SEARCH_SPACE:[
        Argument(('NAME',)),
        DictOption(('-s', '--search_space',), default={},
               help=kConfigFormat[ConfigType.SEARCH_SPACE]['search_space']['description'], show_default=True)
    ],
    ConfigType.PROJECT:[
        Argument(('NAME',)),
        Option(('-p', '--scripts_path',), required=True,
               help=kConfigFormat[ConfigType.PROJECT]['scripts_path']['description']),
        Option(('-m', '--model_config',), required=True,
               help=kConfigFormat[ConfigType.PROJECT]['model_config']['description']),
        Option(('-s', '--search_space',), required=True,
               help=kConfigFormat[ConfigType.PROJECT]['search_space']['description']),
        Option(('-o', '--hpo_config',), required=True,
               help=kConfigFormat[ConfigType.PROJECT]['hpo_config']['description']),
        Option(('-g', '--grid_config',), required=True,
               help=kConfigFormat[ConfigType.PROJECT]['grid_config']['description'])
    ]
}


@click.group(name='hpo_config')
@click.pass_context
def hpo_config(ctx):
    """
    Manage configuration for hyperparameter optimization
    """
    from hpogrid.configuration import HPOConfiguration
    ctx.obj = {"class": HPOConfiguration}
    

@click.group(name='grid_config')
@click.pass_context
def grid_config(ctx):
    """
    Manage configuration for grid job submission
    """
    from hpogrid.configuration import GridConfiguration
    ctx.obj = {"class": GridConfiguration}
        
    
@click.group(name='model_config')
@click.pass_context
def model_config(ctx):
    """
    Manage configuration for the machine learning model
    """
    from hpogrid.configuration import ModelConfiguration
    ctx.obj = {"class": ModelConfiguration}
    
@click.group(name='search_space')
@click.pass_context
def search_space_config(ctx):
    """
    Manage configuration for hyperparameter search space'
    """
    from hpogrid.configuration import SearchSpaceConfiguration
    ctx.obj = {"class": SearchSpaceConfiguration}
    
    
@click.group(name='project')
@click.pass_context
def project_config(ctx):
    """
    Manage an HPO project
    """
    from hpogrid.configuration import ProjectConfiguration
    ctx.obj = {"class": ProjectConfiguration}
    

@click.command(name='list')
@click.option("-e", "--expr", help="filter out configuration files that matches the expression")
@click.pass_context
def list_config(ctx, **kwargs):
    """
    List configuration files
    """
    cls = ctx.obj["class"]()
    cls.list(**kwargs)
    
@click.command(name='show', short_help="Display the contents of a configuration file")
@click.argument("NAME")
@click.pass_context
def show_config(ctx, **kwargs):
    """
    Display the contents of a configuration file
    
    NAME: Name of the configuration file
    """    
    cls = ctx.obj["class"]()
    cls.show(**kwargs)    
    
@click.command(name='remove', short_help="Remove configuration file(s)")
@click.argument("NAME")
@click.pass_context
def remove_config(ctx, **kwargs):
    """
    Remove configuration file(s)
    
    NAME: Name of the configuration file (accept wildcard)
    """ 
    
    cls = ctx.obj["class"]()
    cls.remove(**kwargs)
    
@click.command(name='where', short_help="Display the path of a configuration file")
@click.argument("NAME")
@click.pass_context
def locate_config(ctx, **kwargs):
    """
    Display the path of a configuration file
    
    NAME: Name of the configuration file
    """    
    cls = ctx.obj["class"]()
    cls.locate(**kwargs)
    
@click.pass_context 
def create_config(ctx, **kwargs):
    """
    Create a configuration file
    
    NAME: Name of the configuration file
    """
    cls = ctx.obj["class"]()
    cls.create(**kwargs)
    cls.save(action=ConfigAction.CREATE)
    
@click.pass_context
def update_config(ctx, **kwargs):
    """
    Update a configuration file
    
    NAME: Name of the configuration file
    """
    cls = ctx.obj["class"]()
    cls.update(**kwargs)
    cls.save(action=ConfigAction.UPDATE)
    
@click.pass_context
def recreate_config(ctx, **kwargs):
    """
    Recreate a configuration file
    
    NAME: Name of the configuration file
    """
    cls = ctx.obj["class"]()
    cls.recreate(**kwargs)
    cls.save(action=ConfigAction.RECREATE)
    
for gp in [hpo_config, grid_config, model_config, search_space_config, project_config]:
    gp.add_command(list_config)
    gp.add_command(show_config)
    gp.add_command(remove_config)
    gp.add_command(locate_config)
    
for gp, params in [(hpo_config, config_params[ConfigType.HPO]),
                   (grid_config, config_params[ConfigType.GRID]),
                   (model_config, config_params[ConfigType.MODEL]),
                   (search_space_config, config_params[ConfigType.SEARCH_SPACE]),
                   (project_config, config_params[ConfigType.PROJECT])]:
    cmd_create = click.Command('create', callback=create_config, params=params, 
                               short_help="Create a configuration file")
    cmd_update = click.Command('update', callback=update_config, params=deepcopy(params),
                               short_help="Update a configuration file")
    cmd_recreate = click.Command('recreate',  callback=recreate_config, params=params,
                                 short_help="Recreate a configuration file")
    for param in cmd_update.params:
        if isinstance(param, click.Option):
            param.required = False
            param.default = None
    gp.add_command(cmd_create)
    gp.add_command(cmd_update)
    gp.add_command(cmd_recreate)


class PBTScheduler():
    
    def __init__(self):
        self.scheduler = None
        
    def create(self, metric, mode, space, **args):
        from hpogrid.search_space import PBTSpace
        search_space = PBTSpace(space).search_space
        from ray.tune.schedulers import PopulationBasedTraining
        self.scheduler = PopulationBasedTraining(
            metric=metric, 
            mode=mode, 
            hyperparam_mutations = search_space,
            **args)
        return self.scheduler


import os
import sys
import json
from typing import List, Dict
import numpy as np

from hpogrid.core.defaults import kDefaultGenerator, kDefaultMetric, kDefaultMetricMode, GeneratorMethod
from hpogrid.utils.helper import NpEncoder
from hpogrid.components import AbstractObject

class SteeringIDDS(AbstractObject):
    
    def __init__(self, verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)

    def get_generator(self, space:Dict, metric:str=kDefaultMetric, mode:str=kDefaultMetricMode,
                      lib:str=kDefaultGenerator, **args):
        _lib = GeneratorMethod.parse(lib)
        if _lib == GeneratorMethod.NEVERGRAD:
            from hpogrid.generators import NeverGradGenerator
            return NeverGradGenerator(space, metric, mode, **args)
        elif _lib == GeneratorMethod.SKOPT:
            from hpogrid.generators import SkOptGenerator
            return SkOptGenerator(space, metric, mode, **args)
        elif _lib == GeneratorMethod.HYPEROPT:
            from hpogrid.generators import HyperOptGenerator
            return HyperOptGenerator(space, metric, mode, **args)
        elif _lib == GeneratorMethod.AX:
            from hpogrid.generators import AxGenerator
            return AxGenerator(space, metric, mode, **args)
        elif _lib == GeneratorMethod.BOHB:
            from hpogrid.generators import BOHBGenerator
            return BOHBGenerator(space, metric, mode, **args) 
        elif _lib == GeneratorMethod.GRID:
            from hpogrid.generators import GridGenerator
            return GridGenerator(space, metric, mode, **args)
        elif _lib == GeneratorMethod.OPTUNA:
            from hpogrid.generators import OptunaGenerator
            return OptunaGenerator(space, metric, mode, **args)         
        else:
            raise ValueError(f'Generator from library {lib} is not supported')

    def validate_idds_input(self, idds_input):
        '''
        old format
        {"points": [[{hyperparameter_point_1}, <loss_or_None>], ..., [{hyperparameter_point_N}, <loss_or_None>]], "opt_space": <opt space>}
        new format
        {"points": [[[<model_id_1>, {hyperparameter_point_1}], <loss_or_None>], ..., [[<model_id_N>, {hyperparameter_point_N}], <loss_or_None>]], "opt_space": ["model_id": <model_id>, "search_space": <search_space>]}
        '''
        keys = ['points', 'opt_space']
        # check main keys
        for key in keys:
            if key not in idds_input:
                raise KeyError('Key {} not found in idds input.'
                    ' Please check idds input format'.format(key))
        if not isinstance(idds_input['points'], List):
            raise ValueError('idds should have the data structure: Dict[List[...]]')

        for point in idds_input['points']:
            if not isinstance(point, List):
                raise ValueError('idds should have the data structure: Dict[List[List[...]]]')
            
            if (not isinstance(point[0], Dict)) and (not (isinstance(point[0], List) and isinstance(point[0][0], int) and isinstance(point[0][1], Dict))):
                raise ValueError('idds should have the data structure: Dict[List[List[Dict, Value]]] or Dict[List[List[List[int, Dict], Value]]]')
        # old format
        if isinstance(idds_input['opt_space'], Dict):
            return 1
        # new format
        elif isinstance(idds_input['opt_space'], List):
            for search_space in idds_input['opt_space']:
                if not isinstance(search_space, Dict):
                    raise ValueError('idds search space should be a dictionary not {}'.format(type(search_space)))
            return 2
        else:
            raise ValueError('invalid format for idds opt_space')

    def parse_idds_input(self, file):
        if not file:
            return None

        with open(file,'r') as input_file:
            idds_data = json.load(input_file)
            
        idds_format = self.validate_idds_input(idds_data)
        
        # old format
        if idds_format == 1:
            points = []
            results = []
            pending = 0
            input_data = idds_data['points']
            for data in input_data:
                point, result = data[0], data[1]
                points.append(point)
                results.append(result)
            search_space = idds_data.get('opt_space', None)
            return (points, results, search_space)
        
        elif idds_format == 2:
            model_results = {}
            for opt_space in idds_data.get('opt_space', []):
                model_id = opt_space.get('model_id', None)
                search_space = opt_space.get('search_space', None)
                if model_id is None:
                    raise ValueError('cannot extract model id from idds opt space')
                model_results[model_id] = {}
                model_results[model_id]['search_space'] = search_space
                model_results[model_id]['points'] = []
                model_results[model_id]['results'] = []
            for data in idds_data.get('points', []):
                model_id = data[0][0]
                if model_id not in model_results:
                    raise ValueError(f'missing search space definition for model id: {model_id}')
                point, result = data[0][1], data[1]
                model_results[model_id]['points'].append(point)
                model_results[model_id]['results'].append(result)
            return model_results
        else:
            raise RuntimeError('Failed to parse idds input')
    
    def generate_points(self, search_space, n_point, metric=kDefaultMetric, mode=kDefaultMetricMode,
                        lib=kDefaultGenerator, max_point=None, points=None,
                        results=None, **args):
        
        if not search_space:
            raise RuntimeError('search space is not specified in either idds input or '
            'through command line argument')
        
        # create generator
        generator = self.get_generator(search_space, metric, mode, lib, **args)

        n_evaluated = 0
        n_pending = 0

        # feed evaluated points
        if points and results:
            generator.feed(points=points, results=results)
            self.stdout.info(f'INFO: Fed the following results to the {lib} optimizer')
            generator.show(points=points, results=results)
            n_pending = results.count(None)
            self.stdout.info(f'INFO: There are {n_pending} point(s) pending.')
            generator.show_pending(points=points, results=results)
            n_evaluated = len(points) 

        # determine number of points to generate
        max_point = 9999 or max_point
        n_remaining = max_point - n_evaluated 
        if n_remaining < 0:
            raise ValueError('there are already more evaluated points than the'
            ' maximum points to generate for idds workflow')
        if n_point - n_pending <= 0:
            self.stdout.info('INFO: There are more points pending than the number of points to generate. '
                             'No new points will be generated.')
            return []
        
        n_generate = min(n_point-n_pending, n_remaining)

        # generate points
        new_points = generator.ask(n_generate)
        if GeneratorMethod.parse(lib) == GeneratorMethod.NEVERGRAD:
            extra_text = f" ({generator.searcher.name})"
        else:
            extra_text = ""
        self.stdout.info(f'INFO: Generated {n_generate} new points using the library {lib}{extra_text}')
        generator.show(new_points)

        return new_points

    def run_generator(self, space, n_point, metric=kDefaultMetric,
                      mode=kDefaultMetricMode, lib=kDefaultGenerator, max_point=None,
                      infile=None, outfile=None, **args):
        
        parsed_data = self.parse_idds_input(infile)
        
        # case no input points
        if parsed_data is None:
            if not space:
                raise ValueError('search space can not be empty')
            with open(space, 'r') as space_input:
                search_space = json.load(space_input)
            new_points = self.generate_points(search_space=search_space, n_point=n_point,
                                             metric=metric, mode=mode,
                                             lib=lib, max_point=max_point, **args)
        # old format
        elif isinstance(parsed_data, tuple):
            points, results, search_space = parsed_data[0], parsed_data[1], parsed_data[2]
            new_points = self.generate_points(search_space=search_space, n_point=n_point,
                                             metric=metric, mode=mode,
                                             lib=lib, max_point=max_point, 
                                             points=points, results=results, **args)
        elif isinstance(parsed_data, dict):
            new_points = []
            for model_id in parsed_data:
                self.stdout.info(f'Starting point generation for model ID: {model_id}')
                data = parsed_data[model_id]
                points, results, search_space = data["points"], data["results"], data["search_space"]
                model_specific_new_points = self.generate_points(search_space=search_space, n_point=n_point,
                                                                metric=metric, mode=mode,
                                                                lib=lib, max_point=max_point, 
                                                                points=points, results=results, **args)
                for point in model_specific_new_points:
                    new_points.append(tuple([model_id, point]))

        # save output
        if outfile:
            with open(outfile, 'w') as out:
                json.dump(new_points, out, indent=2, cls=NpEncoder)


import os
import sys
import glob
import json
import fnmatch
from copy import deepcopy
from json import JSONDecodeError
from typing import Optional, Union, Dict, List

from hpogrid.core.defaults import ConfigAction, ConfigType, kConfigFormat
from hpogrid.core.config_manager import get_config_dir, get_config_list, load_configuration
from hpogrid.utils import stylus

from hpogrid.components import AbstractObject

class ConfigurationBase(AbstractObject):
    
    _CONFIG_TYPE_ = ConfigType.BASE
    _CONFIG_DISPLAY_NAME_ = '<CONFIG>'
    _LIST_COLUMNS_ = ['General Configuration']
    _SHOW_COLUMNS_ = ['Attribute', 'Value']    
    _CONFIG_FORMAT_ = {}
    _OUTPUT_EXPR_ = "{name}"
    _DEFAULT_OPT_ARG_ = '0'*99
    
    def __init__(self, name:Optional[str]=None, verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)
        self._config = {}
        self._name = name
        
    def __getitem__(self, key:str):
        return self.config[key]
        
    @classmethod
    def load(cls, filename:str):
        config = cls.load_config(filename)
        instance = cls()
        instance._config = config
        return instance
        
    @property
    def config(self):
        return self._config
    
    @property
    def name(self):
        '''Name of configuration
        '''
        return self._name
        
    @classmethod
    def get_config_dir(cls):
        return get_config_dir(cls._CONFIG_TYPE_)
    
    @classmethod
    def get_config_path(cls, config_name:str=None, extension:str='json'):
        """Returns the full path of a configuration file
        
        Args:
            config_name: str
                Name of the configuration file
            extension: str, default='json'
                File extension for configuration file
        """
        
        config_dir = cls.get_config_dir()
        filename   = cls._OUTPUT_EXPR_.format(name=config_name)
        filename   = os.path.splitext(filename)[0] + f".{extension}"
        config_path = os.path.join(config_dir, filename)
        return config_path    
    
    @classmethod
    def get_config_list(cls, expr:str=None):
        """Returns a list of configuration files for a specific type of configuration
        
        Args:
            expr: str
                Regular expression for filtering name of configuration files
        """
        return get_config_list(cls._CONFIG_TYPE_, expr)
    
    @classmethod
    def load_config(cls, name:str):
        """Returns a specified configuration file
        
        Args:
            name: str
                Name of configuration file
        """
        if not os.path.exists(name):
            config_path = cls.get_config_path(name)
            if not os.path.exists(config_path):
                raise FileNotFoundError(f'configuration file {config_path} does not exist.')
        else:
            config_path = name
        config = load_configuration(config_path)
        return cls.validate(config)
    
    @classmethod
    def locate(cls, name:str):
        """Display the content of a configuration file
        
        Args:
            name: str
                Name of configuration file        
        """
        config_path = cls.get_config_path(name)
        if not os.path.exists(config_path):
            raise FileNotFoundError(f'no configuration file found with name \"{name}\"')
        print(config_path)
    
    @classmethod
    def list(cls, expr:str=None):
        """List out configuration files for a specific type of configuration as a table
        
        Args:
            expr: str
                Regular expression for filtering name of configuration files        
            exclude: list[str]
                Configuration files to exclude from listing
        """
        config_list = cls.get_config_list(expr)
        table = stylus.create_table(config_list, cls._LIST_COLUMNS_)
        print(table)
        
    @classmethod
    def show(cls, name:str):
        """Display the content of a configuration file
        
        Args:
            name: str
                Name of configuration file        
        """
        config = cls.load_config(name)
        table = stylus.create_formatted_dict(config, cls._SHOW_COLUMNS_, indexed=False)
        print(table)
        
    @classmethod
    def remove(cls, expr:str):
        """Removes a configuration file
        
        Args:
            expr: str
                Name of the configuration file to remove (accept wildcard)
        """
        config_list = cls.get_config_list(expr)
        if not config_list:
            cls.stdout.error(f'ERROR: No configuration file found that matches the epxression "{expr}"')
            return None
        for name in config_list:
            config_path = cls.get_config_path(name)
            if os.path.exists(config_path):
                os.remove(config_path)
                cls.stdout.info(f'INFO: Removed file {config_path}')
            else:
                cls.stdout.error(f'ERROR: Cannot remove file {config_path}. File does not exist.')
            
    def _validate_arguments(self, **args):
        if 'name' not in args:
            raise ValueError('missing argument: config_name')
        return args
    
    @classmethod
    def _validate(cls, config):
        cls.stdout.info(f'INFO: Validating {cls._CONFIG_DISPLAY_NAME_} configuration...')
        result = cls.validate(config)
        cls.stdout.info(f'INFO: Successfully validated {cls._CONFIG_DISPLAY_NAME_} configuration')
        return result
    
    @classmethod
    def validate(cls, config:Dict):
        validated_config = deepcopy(config)
        config_format = kConfigFormat[cls._CONFIG_TYPE_]
        for key in config_format:
            if key in config:
                # check if the value type of the config is correct
                value_type = config_format[key]['type']
                # if attribute type is dict, parse string input as dict
                if ((value_type == dict) or (isinstance(value_type, tuple) and (dict in value_type))) and \
                   isinstance(config[key], str):
                    try:
                        validated_config[key] = json.loads(config[key])
                    except JSONDecodeError:
                        raise RuntimeError(f'ERROR: Cannot decode the value of "{key}" as dictionary. '
                                           'Please check your input.')
                # if attribute type is list, parse string input as list by delimitation            
                elif ((value_type == list) or (isinstance(value_type, tuple) and (list in value_type))) and \
                    isinstance(config[key], str):
                    validated_config[key] = config[key].split(",")
                elif not isinstance(config[key], value_type):
                    if isinstance(value_type, tuple):
                        print_type = stylus.type2str(value_type[0])
                    else:
                        print_type = stylus.type2str(value_type)
                    raise ValueError(f'The value of "{key}" must be of type {print_type}')
                # check if the value of the config is allowed
                if ('choice' in config_format[key]) and (config[key] not in config_format[key]['choice']):
                    raise ValueError('The value of "{}" must be one of the followings: {}'.format(
                                     key, str(config_format[key]['choice']).strip('[]')))
            else:
                if config_format[key]['required']:
                    raise ValueError('The required item "{}" is missing from the configuration'.format(key))
                # fill in default config if not specified
                if 'default' in config_format[key]:
                    cls.stdout.info('INFO: Added the item "{}" with default value {} to the configuration'.format(
                                     key, str(config_format[key]['default'])))
                    validated_config[key] = config_format[key]['default']
        for key in config:
            if key not in config_format:
                raise ValueError(f'unknown item "{key}" found in the configuration') 
        return validated_config        
    
    def create(self, **args):
        args = self._validate_arguments(**args)
        return self._configure(action='create', **args)
        
    def recreate(self, **args):
        args = self._validate_arguments(**args)
        return self._configure(action='recreate', **args)
        
    def update(self, **args):
        args = self._validate_arguments(**args)
        config_name = args['name']
        config_path = self.get_config_path(config_name)
        if not os.path.exists(config_path):
            raise FileNotFoundError(f'cannot update file {config_path}. File does not exist.')        
        old_config = self.load_config(config_name)
        new_config = {k:v for k,v in args.items() if v is not None}
        updated_config = {**old_config, **new_config}
        return self._configure(action='update', **updated_config)
    
    def _configure(self, action:Union[ConfigAction, str], **args):
        config_name = args.pop('name')      
        config = self._validate(args)
        self._name = config_name
        self._config = config
        return config
    
    def save(self, name=None, config=None, action:Union[ConfigAction, str]='create'):
        if (name is None) and (config is None):
            name = self._name
            config = self._config
        if name is None: 
            raise ValueError('configuration name undefined')
        config_dir = self.get_config_dir()
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        config_path = self.get_config_path(name)
        action = ConfigAction.parse(action)
        if (os.path.exists(config_path)) and (action==ConfigAction.CREATE):
            self.stdout.error(f'ERROR: {self._CONFIG_DISPLAY_NAME_.title()} configuration "{config_path}" '
                              'already exists. If you want to overwrite, use "recreate" or "update" instead.')
        else:
            with open(config_path, 'w') as config_file:
                json.dump(config, config_file, indent=2)
            action_str = ConfigAction.to_str(action, tense="past").title()
            self.stdout.info(f'INFO: {action_str} {self._CONFIG_DISPLAY_NAME_} configuration {config_path}')
            self.show(name)


from hpogrid.core.defaults import ConfigType
from hpogrid.configuration import ConfigurationBase

class GridConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = ConfigType.GRID
    _CONFIG_DISPLAY_NAME_ = 'grid'
    _LIST_COLUMNS_ = ['Grid Configuration']